# coding=utf-8
import locale
import time
import threading
import socket

import numpy as np

from pymycobot.close_loop import CloseLoop
from pymycobot.robot_info import _interpret_status_code
from pymycobot.common import ProtocolCode,ProGripper


class Pro450Client(CloseLoop):
    def __init__(self, ip='192.168.0.232', netport=4500, debug=False):
        """
        Args:
            ip     : Server IP address, default '192.168.0.232'
            netport : Socket port number, default is 4500
            debug    : whether show debug info
        """
        super(Pro450Client, self).__init__(debug)
        self.SERVER_IP = ip
        self.SERVER_PORT = netport
        self.sock = self.connect_socket()
        self.lock = threading.Lock()
        self.is_stop = False
        self.sync_mode = True
        self.read_threading = threading.Thread(target=self.read_thread, args=("socket",))
        self.read_threading.daemon = True
        self.read_threading.start()
        self.language, _ = locale.getdefaultlocale()
        if self.language not in ["zh_CN", "en_US"]:
            self.language = "en_US"
        self.max_joint, self.min_joint = 0, 0

    def connect_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.SERVER_IP, self.SERVER_PORT))
        return sock
        
    def _mesg(self, genre, *args, **kwargs):
        read_data = super(Pro450Client, self)._mesg(genre, *args, **kwargs)
        if read_data is None:
            return -1
        elif read_data == 1:
            return 1
        elif read_data == -2:
            return 1
        elif read_data == 0:
            return read_data
        if isinstance(read_data, tuple):
            valid_data, data_len = read_data
        else:
            return -1
        res = []
        if data_len in [8, 12, 14, 16, 26, 60]:
            if data_len == 8 and (genre == ProtocolCode.IS_INIT_CALIBRATION):
                if valid_data[0] == 1:
                    return 1
                n = len(valid_data)
                for v in range(1,n):
                    res.append(valid_data[v])
            elif data_len == 8 and genre == ProtocolCode.GET_DOWN_ENCODERS:
                res = self.bytes4_to_int(valid_data)
            elif data_len == 6 and genre in [ProtocolCode.GET_SERVO_STATUS, ProtocolCode.GET_SERVO_VOLTAGES, ProtocolCode.GET_SERVO_CURRENTS]:
                for i in range(data_len):
                    res.append(valid_data[i])
            elif data_len == 8 and genre == ProtocolCode.TOOL_SERIAL_WRITE_DATA:
                res_list = [i for i in valid_data]
                return res_list
            else:
                for header_i in range(0, len(valid_data), 2):
                    one = valid_data[header_i : header_i + 2]
                    res.append(self._decode_int16(one))
        elif data_len == 2:
            if genre in [ProtocolCode.IS_SERVO_ENABLE]:
                return [self._decode_int8(valid_data[1:2])]
            elif genre in [ProtocolCode.GET_ERROR_INFO]:
                return self._decode_int8(valid_data[1:])
            res.append(self._decode_int16(valid_data))
        elif data_len == 3:
            res.append(self._decode_int16(valid_data[1:]))
        elif data_len == 4:
            if genre == ProtocolCode.COBOTX_GET_ANGLE:
                res = self.bytes4_to_int(valid_data)
            for i in range(1,4):
                res.append(valid_data[i])
        elif data_len == 7:
            error_list = [i for i in valid_data]
            if genre == ProtocolCode.IS_INIT_CALIBRATION:
                res = error_list
            else:
                return error_list
            # for i in error_list:
            #     if i in range(16,23):
            #         res.append(1)
            #     elif i in range(23,29):
            #         res.append(2)
            #     elif i in range(32,112):
            #         res.append(3)
            #     else:
            #         res.append(i)
        elif data_len == 24:
            res = self.bytes4_to_int(valid_data)
        elif data_len == 40:
            i = 0
            while i < data_len:
                if i < 28:
                    res += self.bytes4_to_int(valid_data)
                    i+=4
                else:
                    one = valid_data[i : i + 2]
                    res.append(self._decode_int16(one))
                    i+=2
        elif data_len == 30:
            i = 0
            res = []
            while i < 30:
                if i < 9 or i >= 23:
                    res.append(valid_data[i])
                    i+=1
                elif i < 23:
                    one = valid_data[i : i + 2]
                    res.append(self._decode_int16(one))
                    i+=2
        elif data_len == 38:
            i = 0
            res = []
            while i < data_len:
                if i < 10 or i >= 30:
                    res.append(valid_data[i])
                    i+=1
                elif i < 38:
                    one = valid_data[i : i + 2]
                    res.append(self._decode_int16(one))
                    i+=2
        # elif data_len == 56:
        #     for i in range(0, data_len, 8):
                
        #         byte_value = int.from_bytes(valid_data[i:i+4], byteorder='big', signed=True)
        #         res.append(byte_value)
        elif data_len in [6, 9, 32]:
            for i in valid_data:
                res.append(i)
        else:
            if genre in [
                ProtocolCode.GET_SERVO_VOLTAGES,
                ProtocolCode.GET_SERVO_STATUS,
                ProtocolCode.GET_SERVO_TEMPS,
            ]:
                for i in range(data_len):
                    data1 = self._decode_int8(valid_data[i : i + 1])
                    res.append(0xFF & data1 if data1 < 0 else data1)
            res.append(self._decode_int8(valid_data))
        if res == []:
            return -1
        
        if genre in [
            ProtocolCode.ROBOT_VERSION,
            ProtocolCode.GET_ROBOT_ID,
            ProtocolCode.IS_POWER_ON,
            ProtocolCode.IS_CONTROLLER_CONNECTED,
            ProtocolCode.IS_PAUSED,  # TODO have bug: return b''
            ProtocolCode.IS_IN_POSITION,
            ProtocolCode.IS_MOVING,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.IS_ALL_SERVO_ENABLE,
            ProtocolCode.GET_SERVO_DATA,
            ProtocolCode.GET_DIGITAL_INPUT,
            ProtocolCode.GET_GRIPPER_VALUE,
            ProtocolCode.IS_GRIPPER_MOVING,
            ProtocolCode.GET_SPEED,
            ProtocolCode.GET_ENCODER,
            ProtocolCode.GET_BASIC_INPUT,
            ProtocolCode.GET_TOF_DISTANCE,
            ProtocolCode.GET_END_TYPE,
            ProtocolCode.GET_MOVEMENT_TYPE,
            ProtocolCode.GET_REFERENCE_FRAME,
            ProtocolCode.GET_FRESH_MODE,
            ProtocolCode.GET_GRIPPER_MODE,
            ProtocolCode.SET_SSID_PWD,
            ProtocolCode.GET_ERROR_DETECT_MODE,
            ProtocolCode.POWER_ON,
            ProtocolCode.POWER_OFF,
            ProtocolCode.RELEASE_ALL_SERVOS,
            ProtocolCode.RELEASE_SERVO,
            ProtocolCode.FOCUS_ALL_SERVOS,
            ProtocolCode.FOCUS_SERVO,
            ProtocolCode.STOP,
            ProtocolCode.SET_BREAK,
            ProtocolCode.IS_BTN_CLICKED,
            ProtocolCode.GET_CONTROL_MODE,
            ProtocolCode.GET_VR_MODE,
            ProtocolCode.GET_FILTER_LEN,
            ProtocolCode.IS_SERVO_ENABLE,
            ProtocolCode.GET_POS_SWITCH,
            ProtocolCode.GET_TOOL_MODIFY_VERSION,
            ProtocolCode.GET_FUSION_PARAMETERS,
            ProtocolCode.GET_MAX_ACC,
            ProtocolCode.GET_ERROR_INFO,
            ProtocolCode.GET_COLLISION_MODE,
            ProtocolCode.GET_IDENTIFY_MODE,
        ]:
            return self._process_single(res)
        elif genre in [ProtocolCode.GET_ANGLES]:
            return [self._int2angle(angle) for angle in res]
        elif genre in [
            ProtocolCode.GET_COORDS,
            ProtocolCode.MERCURY_GET_BASE_COORDS,
            ProtocolCode.GET_TOOL_REFERENCE,
            ProtocolCode.GET_WORLD_REFERENCE,
        ]:
            if res:
                r = []
                for idx in range(3):
                    r.append(self._int2coord(res[idx]))
                for idx in range(3, 6):
                    r.append(self._int2angle(res[idx]))
                return r
            else:
                return res
        elif genre in [ProtocolCode.GET_SERVO_VOLTAGES]:
            return [self._int2coord(angle) for angle in res]
        elif genre in [ProtocolCode.SOLVE_INV_KINEMATICS]:
            if res == [-57295, -57295, -57295, -57295, -57295, -57295]:
                return 'No solution for conversion'
            return [self._int2angle(angle) for angle in res]
        elif genre in [ProtocolCode.GET_BASIC_VERSION, ProtocolCode.SOFTWARE_VERSION, ProtocolCode.GET_ATOM_VERSION]:
            return self._int2coord(self._process_single(res))
        elif genre in [
            ProtocolCode.GET_JOINT_MAX_ANGLE,
            ProtocolCode.GET_JOINT_MIN_ANGLE,
        ]:
            return self._int2coord(res[0])
        elif genre == ProtocolCode.GET_ANGLES_COORDS:
            r = []
            for index in range(len(res)):
                if index < 7:
                    r.append(self._int2angle(res[index]))
                elif index < 10:
                    r.append(self._int2coord(res[index]))
                else:
                    r.append(self._int2angle(res[index]))
            return r
        elif genre == ProtocolCode.GO_ZERO:
            r = []
            if res:
                if 1 not in res[1:]:
                    return res[0]
                else:
                    for i in range(1, len(res)):
                        if res[i] == 1:
                            r.append(i)
            return r
        elif genre in [ProtocolCode.COBOTX_GET_ANGLE, ProtocolCode.COBOTX_GET_SOLUTION_ANGLES, ProtocolCode.GET_POS_OVER]:
            return self._int2angle(res[0])
        elif genre == ProtocolCode.MERCURY_ROBOT_STATUS:
            if len(res) == 32:
                parsed = res[:8]
                for start in (8, 20):
                    for i in range(start, start + 12, 2):
                        val = (res[i] << 8) | res[i + 1]
                        parsed.append(self._val_to_bits_list(val))

                return parsed
        elif genre == ProtocolCode.IS_INIT_CALIBRATION:
            if res == [1] * 7:
                return 1
            return res
        elif genre == ProtocolCode.GET_BASE_EXTERNAL_CONFIG:
            mode = res[0]
            baud_rate = int.from_bytes(res[1:5], byteorder="little", signed=False)
            timeout = int.from_bytes(res[5:9], byteorder="little", signed=False)
            return [mode, baud_rate, timeout]
        else:
            return res

    def _val_to_bits_list(self, val):
        if val == 0:
            return 0
        return [bit for bit in range(16) if (val >> bit) & 1]

    def _modbus_crc(self, data: bytes) -> bytes:
        crc = 0xFFFF
        for byte in data:
            crc ^= byte
            for _ in range(8):
                if crc & 0x0001:
                    crc >>= 1
                    crc ^= 0xA001
                else:
                    crc >>= 1
        return crc.to_bytes(2, byteorder='little')

    def _send_modbus_command(self, gripper_id, func_code, reg_addr, value_high=None, value_low=None, expect_len=7):
        """
        General Modbus command sending method

        Args:
            gripper_id: Device ID
            func_code: Function code (0x03 = read, 0x06 = write)
            reg_addr: Register address
            value_high: High byte of the write data (can be None for read operations)
            value_low: Low byte of the write data (can be None for read operations)
            expect_len: Expected return data length
        """
        cmd = [gripper_id, func_code, (reg_addr >> 8) & 0xFF, reg_addr & 0xFF]
        if func_code == 0x06:
            cmd.extend([value_high, value_low])
        else:
            cmd.extend([0x00, 0x00])

        cmd.extend(self._modbus_crc(cmd))
        recv = self.tool_serial_write_data(cmd)
        # time.sleep(0.05)
        # recv = []
        # for _ in range(3):
        #     recv = self.tool_serial_read_data(expect_len)
        #     if recv:
        #         break
        if not recv:
            return cmd, -1
        return cmd, recv

    def _check_gripper_id(self, gripper_id):

        self.calibration_parameters(class_name=self.__class__.__name__, gripper_id=gripper_id)

    def _write_and_check(self, gripper_id, reg_addr, value):
        """Write register and verify return"""
        self._check_gripper_id(gripper_id)
        high, low = (value >> 8) & 0xFF, value & 0xFF
        cmd, recv = self._send_modbus_command(gripper_id, 0x06, reg_addr, high, low)
        if isinstance(recv, (list, bytearray)) and len(recv) >= 6 and recv[1] == 0x06:
            if recv[4] == 0x00 and recv[5] == 0x01:
                return 1
            elif recv[4] == 0x00 and recv[5] == 0x00:
                return 0
            else:
                return -1
        return -1

    def _read_register(self, gripper_id, reg_addr):
        """Reads a register and returns an integer value"""
        self._check_gripper_id(gripper_id)
        _, recv = self._send_modbus_command(gripper_id, 0x03, reg_addr)
        if isinstance(recv, (list, bytearray)) and len(recv) >= 5 and recv[1] == 0x03:
            return (recv[4] << 8) | recv[5]
        return -1

    def _joint_limit_init(self):
        max_joint = np.zeros(6)
        min_joint = np.zeros(6)
        for i in range(6):
            max_joint[i] = self.get_joint_max_angle(i + 1)
            min_joint[i] = self.get_joint_min_angle(i + 1)
        return max_joint, min_joint

    def _joint_limit_judge(self, angles):
        offset = 3
        try:
            for i in range(6):
                if self.min_joint[i] + offset < angles[i] < self.max_joint[i] - offset:
                    pass
                else:
                    if self.language == "zh_CN":
                        return f"当前角度为{angles[i]}, 角度范围为： {self.min_joint[i]} ~ {self.max_joint[i]}"
                    return f"current value = {angles[i]}, limit is {self.min_joint[i]} ~ {self.max_joint[i]}"
        except TypeError:
            return "joint limit error"
        return "over limit error {}".format(angles)

    def _Singularity(self, angles):
        try:
            # Joint 6: 0 and 180 degrees are singular points
            singular_angles = [0, 180]
            state = ""
            offset = 5
            for singular in singular_angles:
                if singular - offset < angles[5] < singular + offset:
                    if self.language == "zh_CN":
                        return f"在关节 6 处检测到奇点：{angles[5]} 度"
                    return f"Singularity detected at joint 6: {angles[5]} degrees"
            return state
        except:
            return "Singularity error"

    def _check_coords(self, new_coords, is_print=0):
        try:
            first_three = new_coords[:3]
            first_three[2] -= 83.64
            info = ""
            # Calculate the Euclidean norm (magnitude)
            magnitude = np.linalg.norm(first_three)
            if is_print == 1:
                if self.language == "zh_CN":
                    info += f"当前臂展为{magnitude}, 最大的臂展为{self.arm_span}"
                else:
                    info += f"Arm span is {magnitude}, max is {self.arm_span}"

            # if magnitude > self.arm_span - 10:
            #     if self.language == "zh_CN":
            #         info += f"当前臂展为{magnitude}超出物理限位, 最大的臂展为{self.arm_span}"
            #     else:
            #         info += f"Arm span is {magnitude} exceeds physical limit, max is {self.arm_span}"
            return info
        except:
            return "check coords error"

    def _status_explain(self, status):
        error_info = _interpret_status_code(self.language, status)
        if error_info != "":
            self.arm_span = 440
        if 0x00 < status <= 0x07:
            angles = self.get_angles()
            if type(self.max_joint) == int and self.max_joint == 0:
                self.max_joint, self.min_joint = self._joint_limit_init()
            error_info += self._joint_limit_judge(angles)
        elif status in [32, 33]:
            error_coords = self.get_coords()
            error_info += self._check_coords(error_coords, 1)
        elif status == 36:
            angles = self.get_angles()
            error_info += self._Singularity(angles)

        return error_info

    def open(self):
        self.sock = self.connect_socket()
        
    def close(self):
        self.sock.close()

    def set_motor_enabled(self, joint_id, state):
        """Set the robot torque state.

        Args:
            joint_id: joint id 1-6, 254-all joints
            state: 1 - enable, 0 - disable
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, set_motor_enabled=joint_id, state=state)
        return self._mesg(ProtocolCode.SET_MOTOR_ENABLED, joint_id, state)

    def set_over_time(self, timeout=1000):
        """
        Set the timeout (unit: ms)
        Default is 1000ms (1 second)

        Args:
            timeout (int): Timeout period, in ms, range 0~65535
        """
        if not isinstance(timeout, int) or not (0 <= timeout <= 65535):
            raise ValueError("Timeout must be an integer between 0 and 65535 milliseconds.")

        high_byte = (timeout >> 8) & 0xFF
        low_byte = timeout & 0xFF

        return self._mesg(ProtocolCode.SET_OVER_TIME, high_byte, low_byte)

    def flash_tool_firmware(self, main_version, modified_version=0):
        """Burn tool firmware

        Args:
            main_version (str): Tool firmware version (format: 'x.y')
            modified_version (int): Tool firmware modified version, 0~255, defaults to 0

        """
        self.calibration_parameters(class_name=self.__class__.__name__, tool_main_version=main_version, tool_modified_version=modified_version)
        main_version = int(float(main_version) *10)
        return self._mesg(ProtocolCode.FLASH_TOOL_FIRMWARE, [main_version], modified_version)

    def get_comm_error_counts(self, joint_id):
        """Read the number of communication exceptions

        Args:
            joint_id (int): joint ID, 1 ~ 6
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.MERCURY_ERROR_COUNTS, joint_id)

    def set_break(self, joint_id, value):
        """Set break point

        Args:
            joint_id: int. joint id 1 - 6
            value: int. 0 - disable, 1 - enable

        Return:
            0 : failed
            1 : success
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, value=value)
        return self._mesg(ProtocolCode.SET_BREAK, joint_id, value)

    def get_tool_modify_version(self):
        """Read end correction version number
        """
        return self._mesg(ProtocolCode.GET_TOOL_MODIFY_VERSION)

    def get_pro_gripper_firmware_version(self, gripper_id=14):
        """ Read the firmware major and minor version numbers

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            version number (float): x.x
        """
        val = self._read_register(gripper_id, ProGripper.MODBUS_GET_FIRMWARE_VERSION)
        return val / 10.0 if val >= 0 else -1

    def get_pro_gripper_firmware_modified_version(self, gripper_id=14):
        """ Read the firmware revision number

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            version number (int)
        """
        val = self._read_register(gripper_id, ProGripper.MODBUS_GET_FIRMWARE_MODIFY_VERSION)
        return val if val >= 0 else -1

    def set_pro_gripper_id(self, target_id, gripper_id=14):
        """ Set the gripper ID

        Args:
            target_id (int): Target ID, 1 ~ 254
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, target_id=target_id)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_ID, target_id)

    def get_pro_gripper_id(self, gripper_id=14):
        """ Read the gripper ID

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            gripper_id (int): 1 ~ 254
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_ID)

    def set_pro_gripper_angle(self, gripper_angle, gripper_id=14):
        """ Set the gripper angle

        Args:
            gripper_angle (int): 0 ~ 100
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_angle=gripper_angle)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_ANGLE, gripper_angle)

    def get_pro_gripper_angle(self, gripper_id=14):
        """ Get the gripper angle

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            gripper_angle (int): 0 ~ 100
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_ANGLE)

    def set_pro_gripper_open(self, gripper_id=14):
        """ Open the gripper

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        return self.set_pro_gripper_angle(100, gripper_id)

    def set_pro_gripper_close(self, gripper_id=14):
        """ Close the gripper

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        return self.set_pro_gripper_angle(0, gripper_id)

    def set_pro_gripper_calibration(self, gripper_id=14):
        """ Set the gripper zero position

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_CALIBRATION, 0)

    def get_pro_gripper_status(self, gripper_id=14):
        """ Get the gripper status

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            0 - Moving
            1 - Stopped moving, no clamping detected
            2 - Stopped moving, clamping detected
            3 - After clamping detected, the object fell
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_STATUS)

    def set_pro_gripper_enabled(self, state, gripper_id=14):
        """ Set the gripper enable state

        Args:
            state (bool): 0 or 1, 0 - Disable 1 - Enable
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, state=state)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_ENABLED, state)

    def set_pro_gripper_torque(self, gripper_torque, gripper_id=14):
        """ Set the gripper torque

        Args:
            gripper_torque (int): 0 ~ 100
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_torque=gripper_torque)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_TORQUE, gripper_torque)

    def get_pro_gripper_torque(self, gripper_id=14):
        """ Set the gripper torque

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            gripper_torque (int): 0 ~ 100
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_TORQUE)

    def set_pro_gripper_speed(self, speed, gripper_id=14):
        """ Set the gripper torque

        Args:
            speed (int): 1 ~ 100
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, speed=speed)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_SPEED, speed)

    def get_pro_gripper_speed(self, gripper_id=14):
        """ Get the gripper speed

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Return:
            speed (int): 1 ~ 100
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_SPEED)

    def set_pro_gripper_abs_angle(self, gripper_angle, gripper_id=14):
        """ Set the gripper absolute angle

        Args:
            gripper_angle (int): 0 ~ 100
            gripper_id (int): 1 ~ 254, defaults to 14
        """
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_angle=gripper_angle)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_ABS_ANGLE, gripper_angle)

    def set_pro_gripper_io_open_angle(self, gripper_angle, gripper_id=14):
        """ Set the gripper IO open angle

        Args:
            gripper_angle (int): 0 ~ 100
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_angle=gripper_angle)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_IO_OPEN_ANGLE, gripper_angle)

    def get_pro_gripper_io_open_angle(self, gripper_id=14):
        """ Get the gripper IO open angle

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            angle (int): 0 ~ 100
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_IO_OPEN_ANGLE)

    def set_pro_gripper_io_close_angle(self, gripper_angle, gripper_id=14):
        """ Set the gripper IO close angle

        Args:
            gripper_angle (int): 0 ~ 100
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, gripper_angle=gripper_angle)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_IO_CLOSE_ANGLE, gripper_angle)

    def get_pro_gripper_io_close_angle(self, gripper_id=14):
        """ Get the gripper IO close angle

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            angle (int): 0 ~ 100
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_IO_CLOSE_ANGLE)

    def set_pro_gripper_mini_pressure(self, pressure_value, gripper_id=14):
        """ Set the gripper mini pressure

        Args:
            pressure_value (int): 0 ~ 254
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, pressure_value=pressure_value)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_MINI_PRESSURE, pressure_value)

    def get_pro_gripper_mini_pressure(self, gripper_id=14):
        """ Get the gripper mini pressure

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            mini pressure (int): 0 ~ 254
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_MINI_PRESSURE)

    def set_pro_gripper_protection_current(self, current_value, gripper_id=14):
        """ Set the gripper protection current

        Args:
            current_value (int): 100 ~ 300
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            1 - success, 0 - failed
        """
        self.calibration_parameters(class_name=self.__class__.__name__, current_value=current_value)
        return self._write_and_check(gripper_id, ProGripper.MODBUS_SET_PROTECTION_CURRENT, current_value)

    def get_pro_gripper_protection_current(self, gripper_id=14):
        """ Get the gripper protection current

        Args:
            gripper_id (int): 1 ~ 254, defaults to 14

        Returns:
            current_value (int): 100 ~ 300
        """
        return self._read_register(gripper_id, ProGripper.MODBUS_GET_PROTECTION_CURRENT)

    def set_fresh_mode(self, mode):
        """Set command refresh mode

        Args:
            mode: int.
                1 - Always execute the latest command first.
                0 - Execute instructions sequentially in the form of a queue.
        """
        self.calibration_parameters(class_name=self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.SET_FRESH_MODE, mode)

    def get_fresh_mode(self):
        """Query sports mode

        Returns:
            0 - interpolation mode, 1 - refresh mode
        """
        return self._mesg(ProtocolCode.GET_FRESH_MODE, has_reply=True)

    def set_base_io_output(self, pin_no, pin_signal):
        """Set the base output IO status

        Args:
            pin_no: pin port number. range 1 ~ 12
            pin_signal: 0 - low. 1 - high.
        """
        self.calibration_parameters(class_name=self.__class__.__name__, pin_no_base=pin_no, pin_signal=pin_signal)
        return self._mesg(ProtocolCode.SET_BASIC_OUTPUT, pin_no, pin_signal)

    def get_base_io_input(self, pin_no):
        """Get the input IO status of the base

        Args:
            pin_no: (int) pin port number. range 1 ~ 12
        """
        self.calibration_parameters(class_name=self.__class__.__name__, pin_no_base=pin_no)
        return self._mesg(ProtocolCode.GET_BASIC_INPUT, pin_no)

    def servo_restore(self, joint_id):
        """Abnormal recovery of joints

        Args:
            joint_id (int): Joint ID.
                arm : 1 ~ 6
                All joints: 254
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, servo_restore=joint_id
        )
        return self._mesg(ProtocolCode.SERVO_RESTORE, joint_id)

    def get_angle(self, joint_id):
        """Get single joint angle

        Args:
            joint_id (int): 1 ~ 6.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.COBOTX_GET_ANGLE, joint_id)

    def send_angles(self, angles, speed, _async=False):
        """Send the angles of all joints to robot arm.

        Args:
            angles: a list of angle values(List[float]). len 6.
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, angles=angles, speed=speed)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.SEND_ANGLES, angles, speed, has_reply=True, _async=_async)

    def send_angle(self, joint_id, angle, speed, _async=False):
        """Send one angle of joint to robot arm.

        Args:
            joint_id : Joint id(genre.Angle)， int 1-6.
            angle : angle value(float).
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, angle=angle, speed=speed)
        return self._mesg(ProtocolCode.SEND_ANGLE, joint_id, [self._angle2int(angle)], speed, has_reply=True,
                          _async=_async)

    def send_coord(self, coord_id, coord, speed, _async=False):
        """Send one coord to robot arm.

        Args:
            coord_id (int): coord id, range 1 ~ 6
            coord (float): coord value.
                The coord range of `X` is -466 ~ 466.
                The coord range of `Y` is -466 ~ 466.
                The coord range of `Z` is -230 ~ 614.
                The coord range of `RX` is -180 ~ 180.
                The coord range of `RY` is -180 ~ 180.
                The coord range of `RZ` is -180 ~ 180.
            speed (int): 1 ~ 100
        """

        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=coord_id, coord=coord, speed=speed)
        value = self._coord2int(
            coord) if coord_id <= 3 else self._angle2int(coord)
        return self._mesg(ProtocolCode.SEND_COORD, coord_id, [value], speed, has_reply=True, _async=_async)

    def send_coords(self, coords, speed, _async=False):
        """Send all coords to robot arm.

        Args:
            coords: a list of coords value(List[float]). len 6 [x, y, z, rx, ry, rz]
                The coord range of `X` is -466 ~ 466.
                The coord range of `Y` is -466 ~ 466.
                The coord range of `Z` is -230 ~ 614.
                The coord range of `RX` is -180 ~ 180.
                The coord range of `RY` is -180 ~ 180.
                The coord range of `RZ` is -180 ~ 180.
            speed : (int) 1 ~ 100
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coords=coords, speed=speed)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SEND_COORDS, coord_list, speed, has_reply=True, _async=_async)

    def get_joint_min_angle(self, joint_id):
        """Gets the minimum movement angle of the specified joint

        Args:
            joint_id: Joint id 1 - 6

        Return:
            angle value(float)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MIN_ANGLE, joint_id)

    def get_joint_max_angle(self, joint_id):
        """Gets the maximum movement angle of the specified joint

        Args:
            joint_id: Joint id 1 - 6

        Return:
            angle value(float)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id)
        return self._mesg(ProtocolCode.GET_JOINT_MAX_ANGLE, joint_id)

    def set_joint_max_angle(self, joint_id, degree):
        """Set the maximum angle of the joint (must not exceed the maximum angle specified for the joint)

        Args:
            joint_id (int): Joint id 1 - 6
            degree: The angle range of joint 1 is -165 ~ 165. The angle range of joint 2 is -120 ~ 120. The angle range of joint 3 is -158 ~ 158. The angle range of joint 4 is -165 ~ 165. The angle range of joint 5 is -165 ~ 165. The angle range of joint 6 is -175 ~ 175.

        Return:
            1 - success
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, degree=degree)
        return self._mesg(ProtocolCode.SET_JOINT_MAX, joint_id, degree)

    def set_joint_min_angle(self, joint_id, degree):
        """Set the minimum angle of the joint (must not be less than the minimum angle specified by the joint)

        Args:
            joint_id (int): Joint id 1 - 6.
            degree: The angle range of joint 1 is -165 ~ 165. The angle range of joint 2 is -120 ~ 120. The angle range of joint 3 is -158 ~ 158. The angle range of joint 4 is -165 ~ 165. The angle range of joint 5 is -165 ~ 165. The angle range of joint 6 is -175 ~ 175.

        Return:
            1 - success
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, degree=degree)
        return self._mesg(ProtocolCode.SET_JOINT_MIN, joint_id, degree)

    def set_debug_state(self, log_state):
        """
        Set the debug log mode of the robot.

        Args:
            log_state (int): Debug state as bitmask (0~7)
                0: No debug logs
                1: Only common debug log (_debug.log)
                2: Only motion-related log (_move.log)
                3: Common + motion-related logs (_debug.log+_move.log)
                4: Motor read/control frequency log (_clock_rate_debug.log)
                5: Common + Motor read/control frequency logs (_debug.log+_clock_rate_debug.log)
                6: Motion + Motor read/control frequency logs (_move.log+_clock_rate_debug.log)
                7: All logs

        Returns:
            int: 1-success, 0-failure, -1-error
        """
        self.calibration_parameters(class_name=self.__class__.__name__, log_state=log_state)
        return self._mesg(ProtocolCode.SET_DEBUG_LOG_MODE, log_state)

    def get_debug_state(self):
        """
        Get the current debug log mode of the robot.

        Returns:
            int: Current debug state (0-7), or -1 if failed
                0: No debug logs
                1: Only common debug log (_debug.log)
                2: Only motion-related log (_move.log)
                3: Common + motion-related logs (_debug.log+_move.log)
                4: Motor read/control frequency log (_clock_rate_debug.log)
                5: Common + Motor read/control frequency logs (_debug.log+_clock_rate_debug.log)
                6: Motion + Motor read/control frequency logs (_move.log+_clock_rate_debug.log)
                7: All logs
        """
        return self._mesg(ProtocolCode.GET_DEBUG_LOG_MODE)

    def jog_angle(self, joint_id, direction, speed, _async=True):
        """Jog control angle.

        Args:
            joint_id (int): Joint id 1 - 6.
            direction (int): 0 - decrease, 1 - increase
            speed (int): int range 1 - 100
            _async (bool, optional): Whether to execute asynchronous control. Defaults to True.
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, direction=direction, speed=speed)
        return self._mesg(ProtocolCode.JOG_ANGLE, joint_id, direction, speed, _async=_async, has_reply=True)

    def set_fusion_parameters(self, rank_mode, value):
        """Set speed fusion planning parameters
        Args:
            rank_mode: 0 ~ 4
                0: Restore default parameters (only available in set mode)
                1: Fusion joint velocity
                2: Fusion joint acceleration
                3: Fusion coordinate velocity
                4: Fusion coordinate acceleration
            value: 0 ~ 10000
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, rank_mode=rank_mode, rank_mode_value=value)
        return self._mesg(ProtocolCode.SET_FUSION_PARAMETERS, rank_mode, [value])

    def set_max_acc(self, mode, max_acc):
        """Set maximum acceleration

        Args:
            mode (int): 0 - angle acceleration. 1 - coord acceleration.
            max_acc (int): maximum acceleration value. Angular acceleration range is 1 ~ 200°/s. Coordinate acceleration range is 1 ~ 400mm/s
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, mode=mode, max_acc=max_acc)
        return self._mesg(ProtocolCode.SET_MAX_ACC, mode, [max_acc])

    def jog_increment_angle(self, joint_id, increment, speed, _async=False):
        """Single angle incremental motion control.

        Args:
            joint_id: Joint id 1 - 6.
            increment: Angle increment value
            speed: int (1 - 100)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, joint_id=joint_id, increment_angle=increment, speed=speed)
        scaled_increment = self._angle2int(increment)
        scaled_increment = max(min(scaled_increment, 32767), -32768)
        return self._mesg(ProtocolCode.JOG_INCREMENT, joint_id, [scaled_increment], speed, has_reply=True, _async=_async)

    def jog_increment_coord(self, coord_id, increment, speed, _async=False):
        """Single coordinate incremental motion control.
        This interface is based on a single arm 1-axis coordinate system.

        Args:
            coord_id: axis id 1 - 6.
            increment: Coord increment value
            speed: int (1 - 100)
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, coord_id=coord_id, increment_coord=increment, speed=speed)
        if coord_id <= 3:
            value = self._coord2int(increment)
        else:
            scaled_increment = self._angle2int(increment)
            value = max(min(scaled_increment, 32767), -32768)
        return self._mesg(ProtocolCode.JOG_INCREMENT_COORD, coord_id, [value], speed, has_reply=True, _async=_async)

    def set_communication_mode(self, communication_mode, protocol_mode=None):
        """Set communication mode
        Args:
            communication_mode (int):
                0 - socket communication mode
                1 - 485 communication mode
            protocol_mode (int, optional):
                0 - Custom protocol
                1 - Modbus protocol
                Default: None (means not specified)
        """
        if protocol_mode is not None:
            self.calibration_parameters(
                class_name=self.__class__.__name__, communication_mode=communication_mode, protocol_mode=protocol_mode)
            return self._mesg(ProtocolCode.SET_COMMUNICATION_MODE, communication_mode, protocol_mode)
        else:
            self.calibration_parameters(
                class_name=self.__class__.__name__, communication_mode=communication_mode)
            return self._mesg(ProtocolCode.SET_COMMUNICATION_MODE, communication_mode)

    def get_communication_mode(self):
        """Get communication mode
        Returns:
            communication_mode (int):
                0 - socket communication mode
                1 - 485 communication mode
            protocol_mode (int, optional):
                0 - Custom protocol
                1 - Modbus protocol
        """
        return self._mesg(ProtocolCode.GET_COMMUNICATION_MODE)

    def set_base_external_config(self, communicate_mode, baud_rate, timeout):
        """Bottom external device configuration

        Args:
            communicate_mode (int): 1 - 485. 2 - can
            baud_rate (int): Baud rate
            timeout (int): Timeout ms

        """
        self.calibration_parameters(class_name=self.__class__.__name__, communicate_mode=communicate_mode,
                                    baud_rate=baud_rate, timeout=timeout)
        data = bytearray()
        data += communicate_mode.to_bytes(1, 'big')
        data += baud_rate.to_bytes(4, 'big')
        data += timeout.to_bytes(4, 'big')
        return self._mesg(ProtocolCode.SET_BASE_EXTERNAL_CONFIG, *data)

    def get_base_external_config(self):
        """Read the bottom external device configuration

        Returns:
            communicate_mode (int): 1 - 485. 2 - can
            baud_rate (int): Baud rate
            timeout (int): Timeout ms

        """
        return self._mesg(ProtocolCode.GET_BASE_EXTERNAL_CONFIG)

    def set_model_direction(self, joint_id, direction):
        """Set the direction of the robot model

        Args:
            joint_id (int): joint ID, 1 ~ 6.
            direction (int): 1 - forward, 0 - backward
        """
        return self._mesg(ProtocolCode.SET_MODEL_DIRECTION, joint_id, direction)

    def set_control_mode(self, mode=0):
        """Set robot motion mode

        Args:
            mode (int): 0 - location mode, 1 - torque mode

        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, mode=mode)
        return self._mesg(ProtocolCode.SET_CONTROL_MODE, mode)

    def base_external_can_control(self, can_id, can_data):
        """Bottom external device can control

        Args:
            can_id (int): 1 - 4.
            can_data (list): The maximum length is 64

        """
        self.calibration_parameters(class_name=self.__class__.__name__, can_id=can_id, can_data=can_data)
        can_id_bytes = can_id.to_bytes(4, 'big')
        return self._mesg(ProtocolCode.SET_BASE_EXTERNAL_CONTROL, *can_id_bytes, *can_data)

    def base_external_485_control(self, data):
        """Bottom external device 485 control

        Args:
            data (list): The maximum length is 64

        """
        self.calibration_parameters(class_name=self.__class__.__name__, data_485=data)
        return self._mesg(ProtocolCode.SET_BASE_EXTERNAL_CONTROL, *data)

    def get_error_information(self):
        """Obtaining robot error information

        Return:
            0: No error message.
            1 ~ 6: The corresponding joint exceeds the limit position.
            32-36: Coordinate motion error.
                32: No coordinate solution. Please check if the arm span is near the limit.
                33: No adjacent solution for linear motion.
                34: Velocity fusion error.
                35: No adjacent solution for null space motion.
                36: No solution for singular position. Please use joint control to leave the singular point.
        """
        return self._mesg(ProtocolCode.GET_ERROR_INFO)

    def set_color(self, r=0, g=0, b=0):
        """Set the light color on the top of the robot end.

        Args:
            r (int): 0 ~ 255
            g (int): 0 ~ 255
            b (int): 0 ~ 255

        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, rgb=[r, g, b])
        return self._mesg(ProtocolCode.SET_COLOR_PRO450, r, g, b)

    def parameter_identify(self):
        """Kinetic parameter identification"""
        return self._mesg(ProtocolCode.PARAMETER_IDENTIFY)

    def fourier_trajectories(self, trajectory):
        """Execute dynamic identification trajectory

        Args:
            trajectory (int): 0 ~ 1
        """
        self.calibration_parameters(
            class_name=self.__class__.__name__, trajectory=trajectory)
        return self._mesg(ProtocolCode.FOURIER_TRAJECTORIES, trajectory)

    def set_world_reference(self, coords):
        """Set the world coordinate system

        Args:
            coords: a list of coords value(List[float]). [x(mm), y, z, rx(angle), ry, rz]
        """
        self.calibration_parameters(class_name = self.__class__.__name__, world_coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_WORLD_REFERENCE, coord_list)

    def set_tool_reference(self, coords):
        """Set tool coordinate system

        Args:
            coords: a list of coords value(List[float])
        """
        self.calibration_parameters(class_name = self.__class__.__name__, tool_coords=coords)
        coord_list = []
        for idx in range(3):
            coord_list.append(self._coord2int(coords[idx]))
        for angle in coords[3:]:
            coord_list.append(self._angle2int(angle))
        return self._mesg(ProtocolCode.SET_TOOL_REFERENCE, coord_list)

    def go_home(self, speed=20, _async=False):
        """Control the machine to return to the zero position.

        Args:
            speed (int): 1 ~ 100
        Return:
            1 : All motors return to zero position.
            0 : failed.
        """
        return self.send_angles([0, 0, 0, 0, 0, 0], speed, _async=_async)




#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import serial
import struct
import logging
import logging.handlers


def setup_logging(name: str = __name__, debug: bool = False) -> logging.Logger:
    debug_formatter = logging.Formatter(
        fmt="%(asctime)s.%(msecs)06d %(levelname).4s [%(name)s] %(message)s",
        datefmt="%H:%M:%S",
    )
    logger = logging.getLogger(name)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(debug_formatter)
    if debug is True:
        logger.addHandler(stream_handler)
        logger.setLevel(logging.INFO)  # 100M日志
        file_handler = logging.handlers.RotatingFileHandler(
            filename="python_debug.log", maxBytes=100 * 1024 * 1024, backupCount=1
        )
        file_handler.setFormatter(debug_formatter)
        logger.addHandler(file_handler)

    else:
        logger.setLevel(logging.DEBUG)

    return logger


def setup_serial_connect(port, baudrate, timeout=None):
    serial_api = serial.Serial()
    serial_api.port = port
    serial_api.baudrate = baudrate
    serial_api.timeout = timeout
    serial_api.rts = False
    serial_api.dtr = False
    serial_api.open()
    return serial_api


class CommunicationProtocol(object):
    def write(self, command):
        raise NotImplementedError

    def read(self, size=1):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def open(self):
        raise NotImplementedError

    def is_open(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError


class Utils:
    @classmethod
    def process_data_command(cls, args):
        if not args:
            return []

        processed_args = []
        for index in range(len(args)):
            if isinstance(args[index], list):
                data = cls.encode_int16(args[index])
                processed_args.extend(data)
            else:
                processed_args.append(args[index])

        return cls.flatten(processed_args)

    @classmethod
    def flatten(cls, datas):
        flat_list = []
        for item in datas:
            if not isinstance(item, list):
                flat_list.append(item)
            else:
                flat_list.extend(cls.flatten(item))
        return flat_list

    @classmethod
    def float(cls, number, decimal=2):
        return round(number / 10 ** decimal, 2)

    @classmethod
    def encode_int16(cls, data):
        if isinstance(data, int):
            return [
                ord(i) if isinstance(i, str) else i
                for i in list(struct.pack(">h", data))
            ]
        else:
            res = []
            for v in data:
                t = cls.encode_int16(v)
                res.extend(t)
            return res

    @classmethod
    def decode_int16(cls, data):
        return struct.unpack(">h", data)[0]

    @classmethod
    def crc16_check(cls, command):
        crc = 0xffff
        for index in range(len(command)):
            crc ^= command[index]
            for _ in range(8):
                if crc & 1 == 1:
                    crc >>= 1
                    crc ^= 0xA001
                else:
                    crc >>= 1
        if crc > 0x7FFF:
            crc_res = list(struct.pack(">H", crc))
        else:
            crc_res = cls.encode_int16(crc)

        for i in range(2):
            if isinstance(crc_res[i], str):
                crc_res[i] = ord(crc_res[i])
        return crc_res

    @classmethod
    def crc16_check_bytes(cls, command):
        data = cls.crc16_check(command)
        return bytes(bytearray(data))

    @classmethod
    def get_bits(cls, data):
        reverse_bins = reversed(bin(data)[2:])
        rank = [i for i, e in enumerate(reverse_bins) if e != '0']
        return rank



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import socket
from pymycobot.myagvpro import MyAGVProCommandApi


def setup_socket_connect(host, port, timeout=0.1):
    socket_api = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socket_api.settimeout(timeout)
    socket_api.connect((host, port))
    return socket_api


class MyAGVProSocket(MyAGVProCommandApi):
    def __init__(self, host, port, debug=False, save_serial_log=False):
        super().__init__(debug=debug, save_serial_log=save_serial_log)
        self._socket = setup_socket_connect(host=host, port=port)
        self._serial_filename = 'agvpro_socket_serial.log'
        self._communication_mode = 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __enter__(self):
        return self

    def write(self, command):
        return self._socket.send(command)

    def read(self, size=1):
        try:
            return self._socket.recv(size)
        except socket.timeout:
            return b''

    def close(self):
        self._socket.close()

    def clear(self):
        pass



#!/usr/bin/python

from socket import socket, AF_INET, SOCK_STREAM
import sys
import time
from enum import Enum
import base64
import hashlib
import math
from multiprocessing import Lock
import logging
import numpy as np
from pymycobot.log import setup_logging
from pymycobot.tool_coords import *
from pymycobot.error import MyCobot630ProDataException
import time
from pymycobot.pro630common import Axis, Joint, DI, DO, AI, AO


COORDS_EPSILON = 0.50


class JogMode(Enum):
    JOG_JOINT = 0
    JOG_TELEOP = 1


mutex = Lock()

def parse_mixed_list_string(s):
        s = s.strip()
        if s.startswith("[") and s.endswith("]"):
            try:
                return ast.literal_eval(s)
            except (ValueError, SyntaxError):
                return None
        else:
            try:
                return [int(x.strip()) for x in s.split(",")]
            except ValueError:
                return None


class ElephantRobot(object):
    def __init__(self, host, port, debug=False):
        """Initializes the ElephantRobot instance.

        Args:
            host (str): IP address or hostname of the robot.
            port (int): Port number for the robot's socket server.
            debug (bool): Enables debug mode if True.
        """
        try:
            import numpy as np
        except ImportError:
            raise ImportError("Please install numpy")
        self.debug = debug
        setup_logging(self.debug)
        self.log = logging.getLogger(__name__)
        self.BUFFSIZE = 8 * 1024 * 1024
        self.ADDR = (host, port)
        self.tcp_client = socket(AF_INET, SOCK_STREAM)
        self.is_client_started = False
        tool_reference = np.array([0, 0, 0, 0, 0, 0])
        self.tool_matrix = self.set_tool_reference(tool_reference)

    def get_ip(self):
        """Returns tuple of (IP address or hostname, port) of robot with socket server.

        Returns:
            str: IP address or hostname
        """
        return self.ADDR

    def set_ip(self, host, port):
        """Sets IP address or hostname and port of robot with socket server.
        Cannot set or change IP address or hostname or port of already started client.
        Need to stop client before setting IP address/hostname/port.

        Args:
            host (str): IP address or hostname
            port (int): port

        Returns:
            bool: True if success, False otherwise
        """
        if self.is_client_started:
            print(
                "Error: Cannot change IP if client is started. Stop client to change IP/Host/Port"
            )
            return False
        self.ADDR = (host, port)
        return True

    def start_client(self):
        """Starts the TCP client connection to the robot.

        Returns:
            bool: True if successful, False otherwise.
        """
        try:
            self.tcp_client.connect(self.ADDR)
            self.is_client_started = True
            return True
        except Exception as e:
            print(e)
            return False

    def stop_client(self):
        """Stops the TCP client connection."""
        self.tcp_client.close()
        self.is_client_started = False

    def send_command(self, command):
        """Sends a command to the robot and receives the response.

        Args:
            command (str): Command string to send.

        Returns:
            str: Response from the robot.
        """

        with mutex:
            if not command.endswith('\n'):
                command += '\n'

            self.tcp_client.send(command.encode())

            self.tcp_client.settimeout(10.0)  # Set a 10 second timeout
            try:
                recv_data = self.tcp_client.recv(self.BUFFSIZE).decode()
            except socket.timeout:
                print("Waiting for server response timeout")
                return ""

            res_str = str(recv_data)
            if self.debug:
                print("recv = " + res_str)
            res_arr = res_str.split(":")
            if len(res_arr) == 2:
                return res_arr[1]
            else:
                return ""
    def string_to_coords(self, data):
        """Converts a string representation of coordinates to a list.

        Args:
            data (str): String representation of coordinates.

        Returns:
            list: List of coordinates or invalid coordinates if conversion fails.
        """
        data = data.replace("[", "")
        data = data.replace("]", "")
        data_arr = data.split(",")
        if len(data_arr) == 6:
            try:
                coords_1 = round(float(data_arr[0]), 3)
                coords_2 = round(float(data_arr[1]), 3)
                coords_3 = round(float(data_arr[2]), 3)
                coords_4 = round(float(data_arr[3]), 3)
                coords_5 = round(float(data_arr[4]), 3)
                coords_6 = round(float(data_arr[5]), 3)
                coords = [coords_1, coords_2, coords_3, coords_4, coords_5, coords_6]
                return coords
            except:
                return self.invalid_coords()
        return self.invalid_coords()

    def string_to_double(self, data):
        """Converts a string to a double value.

        Args:
            data (str): String representation of a number.

        Returns:
            float: Converted double value or -9999.99 if conversion fails.
        """
        try:
            val = round(float(data), 3)
            return val
        except:
            return -9999.99

    def float_equal(self, a, b, epsilon=COORDS_EPSILON):
        """Compares 2 floats with given epsilon precision.

        Args:
            a (float): 1st float to compare
            b (float): 2nd float to compare
            epsilon (float): precision (epsilon) to compare

        Returns:
            bool: True if floats are considered equal with given precision,
                  False otherwise
        """
        return math.fabs(a - b) < epsilon

    def coords_equal(self, coords_1, coords_2):
        """Checks if specified coords are equal.

        Args:
            coords_1 (list): first coords to compare
            coords_2 (list): second coords to compare

        Returns:
            bool: True if coords are equal, False otherwise.
        """
        return (
            self.float_equal(coords_1[Axis.X.value], coords_2[Axis.X.value])
            and self.float_equal(coords_1[Axis.Y.value], coords_2[Axis.Y.value])
            and self.float_equal(coords_1[Axis.Z.value], coords_2[Axis.Z.value])
            and self.float_equal(coords_1[Axis.RX.value], coords_2[Axis.RX.value])
            and self.float_equal(coords_1[Axis.RY.value], coords_2[Axis.RY.value])
            and self.float_equal(coords_1[Axis.RZ.value], coords_2[Axis.RZ.value])
        )

    def angles_equal(self, angles_1, angles_2):
        """Checks if specified angles are equal.

        Args:
            angles_1 (list): first angles to compare
            angles_2 (list): second angles to compare

        Returns:
            bool: True if angles are equal, False otherwise.
        """
        return (
            self.float_equal(angles_1[Joint.J1.value], angles_2[Joint.J1.value])
            and self.float_equal(angles_1[Joint.J2.value], angles_2[Joint.J2.value])
            and self.float_equal(angles_1[Joint.J3.value], angles_2[Joint.J3.value])
            and self.float_equal(angles_1[Joint.J4.value], angles_2[Joint.J4.value])
            and self.float_equal(angles_1[Joint.J5.value], angles_2[Joint.J5.value])
            and self.float_equal(angles_1[Joint.J6.value], angles_2[Joint.J6.value])
        )

    def string_to_int(self, data):
        try:
            val = int(data)
            return val
        except:
            return -9999

    def invalid_coords(self):
        """Returns a predefined set of invalid coordinates.

        Returns:
            list: Invalid coordinates.
        """
        coords = [-1.0, -2.0, -3.0, -4.0, -1.0, -1.0]
        return coords

    def detect_robot(self):
        """Detects the robot's model.

        Returns:
            int: Integer value indicating robot model.
        """
        return int(self.get_analog_in(55))

    def start_robot(self):
        """Starts robot.

        Returns:
            bool: currently always returns True
        """
        command = "start_robot()\n"
        res = self.send_command(command)
        return True

    def recover_robot(self):
        """Recovers robot after collision.

        Returns:
            bool: currently always returns True
        """
        return self.start_robot()

    def get_angles(self):
        """Retrieves the current joint angles of the robot.

        Returns:
            list: List of joint angles.
        """
        command = "get_angles()\n"
        res = self.send_command(command)
        return self.string_to_coords(res)

    def get_angle(self, joint: Joint) -> float:
        """Retrieves the angle of a specific joint.

        Args:
            joint (Joint): specific joint.

        Example:
            >>> e.get_angle(Joint.J1)
            >>> 1.234
            >>> e.get_angle(Joint.J2)
            >>> -92.345
            >>> e.get_angle(Joint.J3)
            >>> 3.456
            >>> e.get_angle(Joint.J4)
            >>> -94.567
            >>> e.get_angle(Joint.J5)
            >>> 5.678
            >>> e.get_angle(Joint.J6)
            >>> 6.789

        Returns:
            float: Angle of the specified joint.
        """
        command = "get_angle(" + str(joint) + ")\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def get_coords(self):
        """Retrieves the current coordinates of the robot.

        Returns:
            list: List of coordinates.
        """
        command = "get_coords()\n"
        res = self.send_command(command)
        return self.string_to_coords(res)

    def get_coord(self, axis: Axis) -> float:
        """Retrieves the value of a specific coordinate axis.

        Args:
            axis (Axis): specified Axis.

        Example:
            >>> e.get_coord(Axis.X)
            >>> 0.0
            >>> e.get_coord(Axis.Y)
            >>> 140.0
            >>> e.get_coord(Axis.Z)
            >>> 835.0
            >>> e.get_coord(Axis.RX)
            >>> 90.0
            >>> e.get_coord(Axis.RY)
            >>> 0.0
            >>> e.get_coord(Axis.RZ)
            >>> 180.0

        Returns:
            float: Value of the specified axis.
        """
        command = "get_coord(" + str(axis) + ")\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def get_speed(self):
        """Retrieves the current speed of the robot.

        Returns:
            float: Speed value.
        """
        command = "get_speed()\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def _power_on(self):
        command = "power_on()\n"
        res = self.send_command(command)
        return True

    def _power_off(self):
        command = "power_off()\n"
        res = self.send_command(command)
        return True

    def is_power_on(self):
        """Checks whether robot is powered on.

        Returns:
            bool: True if robot is powered on, False otherwise.
        """
        command = "is_power_on()\n"
        res = self.send_command(command)
        return res == "1"

    def _state_on(self):
        command = "state_on()\n"
        self.send_command(command)

    def _state_off(self):
        command = "state_off()\n"
        self.send_command(command)

    def state_check(self):
        """Checks if robot is enabled.

        Returns:
            bool: True if robot is enabled, False otherwise.
        """
        command = "state_check()\n"
        res = self.send_command(command)
        return res == "1"

    def check_running(self):
        """Checks whether robot is moving.

        Returns:
            bool: True if robot is moving, False otherwise.
        """
        command = "check_running()\n"
        res = self.send_command(command)
        return res == "1"

    def is_in_position(self, coords, jog_mode):
        """Returns True if current position equals passed coords.

        Args:
            coords (list[float]): coords or angles
            jog_mode (JogMode): JogMode enum value

        Returns:
            bool: True if robot is in passed coords, False otherwise
        """
        if type(jog_mode) is not JogMode:
            raise TypeError("jog_mode should be of type JogMode")
        if jog_mode == JogMode.JOG_TELEOP:
            return self.coords_equal(self.get_coords(), coords)
        else:
            return self.angles_equal(self.get_angles(), coords)

    def set_free_move_mode(self, on=True):
        """Enables or disables free move mode.

        Args:
            on (bool, optional): True to enable free move mode,
                                 False to disable free move mode.
                                 Defaults to True.
        """
        command = "set_free_move(" + str(int(on)) + ")\n"
        self.send_command(command)

    def set_payload(self, payload):
        """Sets payload value on the robot.

        Args:
            payload (float): payload value (0.0~2.0)
        """
        command = "set_payload(" + str(payload) + ")\n"
        self.send_command(command)

    def get_payload(self):
        """Returns current payload value.

        Returns:
            float: payload value
        """
        return self.get_analog_out(53)

    def upload_file(self, local_filename, remote_filename):
        """Upload local file to robot via socket. Permissions are checked before
           upload so it cannot overwrite system files (owned by root).

        Args:
            local_filename (str): absolute or relative path to a file to send
            remote_filename (str): absolute or relative file path to where
                                   to upload a file. If path is relative, it is
                                   relative to ~/RoboFlow/upload folder.

        Returns:
            str: ok if success or error message otherwise
        """
        with open(local_filename, "rb") as f:
            content = f.read()
        content_base64 = base64.b64encode(content).decode()
        content_sha256 = hashlib.sha256(content).hexdigest()
        command = "upload_file({},{},{})".format(
            content_base64, remote_filename, content_sha256
        )
        res = self.send_command(command)
        return res

    def read_text_file(self, remote_filename):
        """Reads the content of a text file from the robot.

        Args:
            remote_filename (str): Path to the file on the robot.

        Returns:
            str: Content of the file.
        """
        command = f"read_text_file({remote_filename})"
        res = self.send_command(command)
        return res

    def program_open(self, file_path):
        """Opens a program file on the robot. File must be a g-code file.

        Args:
            file_path (str): Path to the program (g-code) file.

        Returns:
            int: Status code indicating success or failure.
        """
        command = "program_open(" + file_path + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def program_run(self, start_line):
        """Runs a program on the robot starting from a specific line.
        Program must be opened before running using program_open() method.

        Args:
            start_line (int): Line number to start execution.

        Returns:
            int: Status code indicating success or failure.
        """
        command = "program_run(" + str(start_line) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def read_next_error(self):
        """Reads the next error message from the robot.

        Returns:
            str: Error message or an empty string if no errors.
        """
        command = "read_next_error()\n"
        res = self.send_command(command)
        return res

    def clear_all_errors(self):
        """Clears all error messages from the robot."""
        while self.read_next_error() != "":
            pass

    def get_robot_state(self) -> str:
        """Returns current robot state, which is 64 bits represented as a
        string of 64 zeroes or ones.

        Meaning of each bit from eft to right is as follows:
        Bit number | Meaning
        0          | Power on
        1          | Emergency stop button, 0 - pressed, 1 - released
        2          | Joint 6 status, 0 - error, 1 - OK
        3          | Joint 5 status, 0 - error, 1 - OK
        4          | Joint 4 status, 0 - error, 1 - OK
        5          | Joint 3 status, 0 - error, 1 - OK
        6          | Joint 2 status, 0 - error, 1 - OK
        7          | Joint 1 status, 0 - error, 1 - OK
        8          | Reserved
        9          | Reserved
        10         | Reserved
        11         | Motion enabled, 0 - disabled, 1 - enabled
        12         | Hardware Free Move
        13         | Joint 6 servo enabled, 0 - disabled, 1 - enabled
        14         | Joint 5 servo enabled, 0 - disabled, 1 - enabled
        15         | Joint 4 servo enabled, 0 - disabled, 1 - enabled
        16         | Joint 3 servo enabled, 0 - disabled, 1 - enabled
        17         | Joint 2 servo enabled, 0 - disabled, 1 - enabled
        18         | Joint 1 servo enabled, 0 - disabled, 1 - enabled
        19         | Brake activation running, 0 - not running, 1 - running
        20         | Hardware Pause Pressed, 0 - not pressed, 1 - pressed
        21         | Reserved
        22         | IO Run Triggered, 0 - not triggered, 1 - triggered
        23         | IO Stop Triggered, 0 - not triggered, 1 - triggered
        24         | Joint 6 communication status, 0 - error, 1 - OK
        25         | Joint 5 communication status, 0 - error, 1 - OK
        26         | Joint 4 communication status, 0 - error, 1 - OK
        27         | Joint 3 communication status, 0 - error, 1 - OK
        28         | Joint 2 communication status, 0 - error, 1 - OK
        29         | Joint 1 communication status, 0 - error, 1 - OK
        30         | Collision detected, 0 - not detected, 1 - detected
        31         | Reserved
        32         | Reserved
        33         | Reserved
        34         | Reserved
        35         | Reserved
        36         | Reserved
        37         | Reserved
        38         | Reserved
        39         | Reserved
        40         | Reserved
        41         | Reserved
        42         | Reserved
        43         | Reserved
        44         | Reserved
        45         | Reserved
        46         | C Series Physical User button
        47         | C Series Physical Stop button
        48         | C Series Physical Start button
        49         | Reserved
        50         | Reserved
        51         | Reserved
        52         | Reserved
        53         | Reserved
        54         | Reserved
        55         | Reserved
        56         | Reserved
        57         | Reserved
        58         | Reserved
        59         | Reserved
        60         | Reserved
        61         | Reserved
        62         | Reserved
        63         | Reserved

        For example, if robot is powered on, enabled and in position,
        the returned string will be:
        '1011111100000111111000001111110000000000000000110000000000000000'

        Returns:
            str: robot state as a string of 64 zeroes or ones.
        """
        command = "get_robot_state()\n"
        res = self.send_command(command)
        return res

    def set_servos_calibration(self) -> str:
        """Sets servos calibration, that is, current robot pose will become
        robot's zero position.

        This function will shutdown robot. You will need to restart the robot
        and reconnect to continue working with it.

        Returns:
            str: empty string if success, error message otherwise.
        """
        command = "set_servos_calibration()\n"
        res = self.send_command(command)
        return res

    def get_joint_loss_pkg(self, joint_number: Joint) -> int:
        """Retrieves the loss package count for a specific joint.

        Args:
            joint_number (Joint): The joint number for which the loss package
                                  value is requested.

        Returns:
            int: The loss package count for the specified joint.
        """
        command = "get_joint_loss_pkg(" + str(joint_number.value) + ")\n"
        res = self.send_command(command)
        return int(self.string_to_double(res))

    # TODO: rename to set_coords()
    def write_coords(self, coords, speed):
        """Sets the robot's coordinates.

        Args:
            coords (list[float]): List of coordinates [X, Y, Z, RX, RY, RZ].
            speed (int): Speed of movement.
        """
        command = "set_coords("
        for item in coords:
            command += str(round(item, 3)) + ","
        command += str(speed) + ")\n"
        self.send_command(command)

    # TODO: change to set_coord() socket API (already impemented)
    def write_coord(self, axis, value, speed):
        """Sets the value of a specific coordinate axis.

        Args:
            axis (int): Axis index from 0 to 5 (for X, Y, Z, RX, RY, RZ).
            value (float): Value to set for the axis.
            speed (int): Speed of movement.
        """
        coords = self.get_coords()
        if coords != self.invalid_coords():
            coords[axis] = value
            self.write_coords(coords, speed)

    # TODO: rename to set_angles()
    def write_angles(self, angles, speed):
        """Sets the robot's joint angles.

        Args:
            angles (list[float]): List of joint angles [J1, J2, J3, J4, J5, J6].
            speed (int): Speed of movement.
        """
        command = "set_angles("
        for item in angles:
            command += str(round(item, 3)) + ","
        command += str(speed) + ")\n"
        self.send_command(command)

    # TODO: change to set_angle() socket API (already impemented)
    def write_angle(self, joint, value, speed):
        """Sets the angle of a specific joint.

        Args:
            joint (int): Joint index from 0 to 5 (for J1, J2, J3, J4, J5, J6).
            value (float): Angle value to set.
            speed (int): Speed of movement.
        """
        angles = self.get_angles()
        if angles != self.invalid_coords():
            angles[joint] = value
            self.write_angles(angles, speed)

    def set_speed(self, percentage):
        """Sets the robot's movement speed.

        Args:
            percentage (int): Speed percentage (0-100).
        """
        command = "set_speed(" + str(percentage) + ")\n"
        self.send_command(command)

    def set_carte_torque_limit(self, axis_str, value):
        """Sets the torque limit for a specific Cartesian axis.

        Args:
            axis_str (str): Axis name (e.g., "X", "Y", "Z", "RX", "RY", "RZ").
            value (float): Torque limit value.
        """
        command = "set_torque_limit(" + axis_str + "," + str(value) + ")\n"
        self.send_command(command)

    def set_upside_down(self, up_down):
        """Sets the robot's upside-down mode.

        Args:
            up_down (bool): True to enable upside-down mode, False to disable.
        """
        up = "1"
        if up_down:
            up = "0"
        command = "set_upside_down(" + up + ")\n"
        self.send_command(command)

    def task_stop(self):
        """Stops the movement of robot."""
        command = "task_stop()\n"
        self.send_command(command)

    def jog_angle(self, joint_str, direction, speed):
        """Performs a jog operation on a specific joint.

        Args:
            joint_str (str): Joint identifier (e.g., 'J1', 'J2', ..., 'J6').
            direction (int): Direction of movement (-1 for negative, 1 for positive).
            speed (int): Speed of the jog operation.
        """
        command = (
            "jog_angle(" + joint_str + "," + str(direction) + "," + str(speed) + ")\n"
        )
        self.send_command(command)

    def jog_coord(self, axis_str, direction, speed):
        """Performs a jog operation on a specific coordinate axis.

        Args:
            axis_str (str): Axis identifier (e.g., 'X', 'Y', 'Z', 'RX', 'RY', 'RZ').
            direction (int): Direction of movement (-1 for negative, 1 for positive).
            speed (int): Speed of the jog operation.
        """
        command = (
            "jog_coord(" + axis_str + "," + str(direction) + "," + str(speed) + ")\n"
        )
        self.send_command(command)

    def get_digital_in(self, pin_number):
        """Retrieves the state of a digital input pin.

        Args:
            pin_number (int): Pin number (0-63).

        Returns:
            int: State of the pin (0 or 1).
        """
        command = "get_digital_in(" + str(int(pin_number)) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_digital_out(self, pin_number):
        """Retrieves the state of a digital output pin.

        Args:
            pin_number (int): Pin number (0-63).

        Returns:
            int: State of the pin (0 or 1).
        """
        command = "get_digital_out(" + str(int(pin_number)) + ")\n"
        print(command)
        res = self.send_command(command)
        return self.string_to_int(res)

    def set_digital_out(self, pin_number, pin_signal):
        """Sets the state of a digital output pin.

        Args:
            pin_number (int): Pin number (0-63).
            pin_signal (int): Signal to set (0 or 1).
        """
        command = (
            "set_digital_out("
            + str(int(pin_number))
            + ","
            + str(int(pin_signal))
            + ")\n"
        )
        self.send_command(command)

    def get_analog_in(self, pin_number):
        """Returns specified analog input pin value.

        Args:
            pin_number (int): pin number (0-63)

        Returns:
            float: pin value
        """
        command = "get_analog_in(" + str(int(pin_number)) + ")\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def get_analog_out(self, pin_number):
        """Returns specified analog output pin value.

        Args:
            pin_number (int): pin number (0-63)

        Returns:
            float: pin value
        """
        command = "get_analog_out(" + str(int(pin_number)) + ")\n"
        res = self.send_command(command)
        return self.string_to_double(res)

    def set_analog_out(self, pin_number, pin_value):
        """Sets specified analog output pin to given value.

        Args:
            pin_number (int): pin number (0~63).
            pin_value (float): pin value
        """
        command = (
            "set_analog_out(" + str(int(pin_number)) + "," + str(int(pin_value)) + ")\n"
        )
        self.send_command(command)

    def get_joint_current(self, joint_number):
        """Retrieves the current value of a specific joint.

        Args:
            joint_number (int): Joint index from 0 to 5 (for J1, J2, J3, J4, J5, J6).

        Returns:
            float: Current value of the specified joint.
        """
        command = "get_joint_current(" + str(joint_number) + ")\n"
        print(command)
        res = self.send_command(command)
        return self.string_to_double(res)

    def send_feed_override(self, override):
        """Sets the feed rate override percentage.

        Args:
            override (int): Feed rate override percentage (0-100).

        Returns:
            int: Status code indicating success or failure.
        """
        command = "set_feed_rate(" + str(override) + ")\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def get_acceleration(self):
        """Retrieves the current acceleration setting of the robot.

        Returns:
            int: Acceleration value.
        """
        command = "get_acceleration()\n"
        res = self.send_command(command)
        return self.string_to_int(res)

    def set_acceleration(self, acceleration):
        """Sets the acceleration value for the robot.

        Args:
            acceleration (int): Acceleration value to set.
        """
        command = "set_acceleration(" + str(acceleration) + ")\n"
        self.send_command(command)

    def command_wait_done(self):
        """Waits until the current command execution is completed."""
        command = "wait_command_done()\n"
        self.send_command(command)

    def wait(self, seconds):
        """Pauses execution for a specified number of seconds.

        Args:
            seconds (int): Number of seconds to wait.
        """
        command = "wait(" + str(seconds) + ")\n"
        self.send_command(command)

    def assign_variable(self, var_name, var_value):
        """Assigns a value to a variable on the robot.

        Args:
            var_name (str): Name of the variable.
            var_value (str): Value to assign to the variable.
        """
        command = 'assign_variable("' + str(var_name) + '",' + str(var_value) + ")\n"
        self.send_command(command)

    def get_variable(self, var_name):
        """Retrieves the value of a variable from the robot.

        Args:
            var_name (str): Name of the variable.

        Returns:
            str: Value of the variable.
        """
        command = 'get_variable("' + str(var_name) + '")\n'
        return self.send_command(command)

    def jog_relative(self, joint_id, angle, speed, mode):
        """Relative jog.

        Example:
            jog_relative('J1', 5, 600, 1)

        Args:
            joint_id (str): 'J1' - 'J6', 'X', 'Y', 'Z', 'RX', 'RY', 'RZ'
            angle (float): relative angle or coord value to move, can be negative
                           to move to other direction
            speed (int): speed value
            mode (int): 0 (coord mode) or 1 (joint mode)

        Returns:
            str: return data from socket
        """
        command = "jog_increment({},{},{},{})\n".format(joint_id, angle, speed, mode)
        return self.send_command(command)

    def enable_manual_brake_control(self, enable=True):
        """Enables or disables manual brake control.

        Args:
            enable (bool, optional): enable (True) or disable (False) manual
                                     brake control. Defaults to True.
        """
        self.set_digital_out(DO.BRAKE_MANUAL_MODE_ENABLE.value, enable)
        time.sleep(0.05)
        for joint in Joint:
            self.release_joint_brake(joint, False)

    def release_joint_brake(self, joint, release=True):
        """Releases or focuses (enables) specified joint's brake.

        Args:
            joint (Joint): joint Joint.J1 ~ Joint.J6
            release (bool): True to release, False to enable brake. Defaults to True.
        """
        self.set_digital_out(DO(joint.value + DO.J1_BRAKE_RELEASE.value).value, release)

    def is_collision_detected(self):
        """Checks if collision is detected.

        Returns:
            bool: True if collision is detected, False otherwise.
        """
        return self.get_digital_in(DI.COLLISION_DETECTED.value) == 1

    def set_gripper_state(self, state, speed):
        """Sets gripper state.

        Args:
            state (int): gripper state, 0 - open, 1 - close
            speed (int): speed, 1-100

        Returns:
            str: return message
        """
        command = "set_gripper_state(" + str(state) + "," + str(speed) + ")\n"
        return self.send_command(command)

    def set_gripper_value(self, value, speed):
        """Sets gripper open/close state.

        Args:
            value (int): gripper open value, 0-100
            speed (int): speed, 1-100

        Returns:
            str: return message
        """
        command = "set_gripper_value(" + str(value) + "," + str(speed) + ")\n"
        return self.send_command(command)

    def set_gripper_calibrate(self):
        """Sets gripper zero position.

        Returns:
            str: return message
        """
        command = "set_gripper_calibrate()\n"
        return self.send_command(command)

    def set_gripper_enabled(self, enabled):
        """Set gripper enabled.

        Args:
            enabled (int): 1 - enabled, 0 - disabled

        Returns:
            str: return message
        """
        command = "set_gripper_enabled(" + str(enabled) + ")\n"
        return self.send_command(command)

    def set_gripper_mode(self, mode):
        """Sets gripper mode. Default mode is IO, after setting mode 485
           cannot set mode back to IO (need gripper reboot).

        Args:
            mode (int): 0 - 485, 1 - IO

        Returns:
            str: return message
        """
        command = "set_gripper_mode(" + str(mode) + ")\n"
        return self.send_command(command)

    def init_ele_gripper(self):
        """Initializes Electric Gripper.

        Returns:
            str: return message
        """
        command = "init_ele_gripper()\n"
        return self.send_command(command)

    def set_ele_gripper_open(self):
        """Fully opens Electric Gripper.

        Returns:
            str: return message
        """
        command = "set_ele_gripper_open()\n"
        return self.send_command(command)

    def set_ele_gripper_close(self):
        """Fully closes Electric Gripper.

        Returns:
            str: return message
        """
        command = "set_ele_gripper_close()\n"
        return self.send_command(command)

    def set_tool_reference(self, tool_reference):
        """Set tool coordinate system.

        Args:
            tool_reference (list):

        Returns:
            _type_: _description_
        """
        rotation_matrix = np.eye(3)  # No rotation example
        translation_vector = tool_reference[:3]  # Tool offset along z-axis of flange
        rotation_matrix = CvtEulerAngleToRotationMatrix(
            tool_reference[3:6] * np.pi / 180.0
        )
        return transformation_matrix_from_parameters(
            rotation_matrix, translation_vector
        )

    def get_tool_coords(self):
        """Set tool coordinate system.

        Returns:
            _type_: _description_
        """
        current_coords = np.array(self.get_coords())
        tool_coords = flangeToTool(current_coords, self.tool_matrix)
        return tool_coords

    def write_tool_coords(self, tool_coords, speed):
        """Tool coordinate motion control

        Args:
            tool_coords (list): _description_
            speed (int): _description_
        """
        flange_coords = toolToflange(tool_coords, self.tool_matrix)
        self.write_coords(flange_coords, speed)

    #力控#
    def force_get_firmware(self, ID):
        """Obtain the main version number of the power control system

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return information
        """
        command = "force_GetFirmware(" + str(ID) + ")\n"
        value = int(self.send_command(command))
        if value == 255:
            return value
        return (int(self.send_command(command)))/10
    
    def force_get_modified(self, ID):
        """Obtain the updated version number of Power Control

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return information
        """
        command = "force_GetModified(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_set_id(self, ID, value):
        """Set force control ID

        Args:
            ID (int): Robot arm ID.
            Value (int): ID value (1~254).
        
        Returns:
            str: return message
        """
        if value < 1 or value > 254:
            raise MyCobot630ProDataException(
            "The Force_SetId value must be 1 ~ 254, but received {}".format(value))
        command = "force_SetGripperId(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_get_id(self, ID):
        """Obtain the ID of the force control machine

        Args:
            ID (int): Robot arm ID.
        
         Returns:
            str: return ID value
        """
        command = "force_GetGripperId(" + str(ID) + ")\n"
        return int(self.send_command(command))

    def force_set_enabled(self, ID, value):
        """Set force control enable

        Args:
            ID (int): Robot arm ID.
            Value (int): Enable value (0 or 1).

        Returns:
            str: return message
        """
        if value < 0 or value > 1:
            raise MyCobot630ProDataException(
            "The Force_SetEnabled value must be 0 ~ 1, but received {}".format(value))
        command = "force_SetGripperEnabled(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_set_angle(self, ID, value):
        """Set force control angle

        Args:
            ID (int): Robot arm ID.
            Value (int): Angle value (0~100).
        
        Returns:
            str: return message
        """
        if value < 0 or value > 100:
            raise MyCobot630ProDataException(
            "The Force_SetAngle value must be 0 ~ 100, but received {}".format(value))
        command = "force_SetAngle(" + str(ID) + "," + str(value) + ")\n"
        # print(command)
        return int(self.send_command(command))

    def force_get_angle(self, ID):
        """Obtain force control angle

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return angle value
        """
        command = "force_GetGripperAngle(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_set_calibrate(self, ID):
        """Initialize force control

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return message
        """
        command = "force_SetGripperCalibrate(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_get_gripper(self, ID):
        """Obtain the motion status of force control

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return status value
        """
        command = "force_GetGripper(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_set_torque(self, ID, value):
        """Set the torque value for force control

        Args:
            ID (int): Robot arm ID.
            Value (int): Torque value (0~100).
        Returns:
            str: return message
        """
        if value < 0 or value > 100:
            raise MyCobot630ProDataException(
            "The Force_SetTorque value must be 0 ~ 100, but received {}".format(value))
        command = "force_SetGripperTorque(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_get_torque(self, ID):
        """Obtain force control torque

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return torque value
        """
        command = "force_GetGripperTorque(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_set_open(self, ID, value):
        """Set the opening angle of the force control IO

        Args:
            ID (int): Robot arm ID.
            Value (int): Angle value (0~100).
        
        Returns:
            str: return message
        """
        if value < 0 or value > 100:
            raise MyCobot630ProDataException(
            "The Force_SetOpen value must be 0 ~ 100, but received {}".format(value))
        command = "force_SetOpen(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_set_close(self, ID, value):
        """Set the closing angle of the force control IO

        Args:
            ID (int): Robot arm ID.
            Value (int): Angle value (0~100).
        
        Returns:
            str: return angle value
        """
        if value < 0 or value > 100:
            raise MyCobot630ProDataException(
            "The Force_SetClose value must be 0 ~ 100, but received {}".format(value))
        command = "force_SetClose(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_set_speed(self, ID, value):
        """Set force control motion speed

        Args:
            ID (int): Robot arm ID.
            Value (int): Speed value (1~100).

        Returns:
            str: return message
        """
        if value < 1 or value > 100:
            raise MyCobot630ProDataException(
            "The Force_SetSpeed value must be 1 ~ 100, but received {}".format(value))
        command = "force_SetSpeed(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_get_speed(self, ID):
        """Obtain force controlled motion speed

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return speed value
        """
        command = "force_GetSpeed(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_get_open(self, ID):
        """Obtain the opening angle of the force control IO

        Args:
            ID (int): Robot arm ID.
            Value (int): Angle value (0~100).
        
        Returns:
            str: return angle value
        """
        command = "force_GetOpen(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_get_close(self, ID):
        """ Obtain the closing angle of the force control IO

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return angle value
        """
        command = "force_GetClose(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_set_absangle(self, ID, value):
        """Set the absolute angle of force control

        Args:
            ID (int): Robot arm ID.
            Value (int): Angle value (0~100).
        
        Returns:
            str: return message
        """
        if value < 0 or value > 100:
            raise MyCobot630ProDataException(
            "The SetAbsAngle  value must be 0 ~ 100, but received {}".format(value))
        command = "force_SetAbsAngle(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def force_pause(self, ID):
        """Pause absolute instruction queue

        Args:
           ID (int): Robot arm ID.
        
        Returns:
            str: return message
        """
        command = "force_Pause(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_resume(self, ID):
        """Open absolute command queue

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return message
        """
        
        command = "force_Resume(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def force_stop(self, ID):
        """Set absolute command to stop

        Args:
            ID (int): Robot arm ID.
        """
        command = "force_Stop(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def froce_get_queuecount(self, ID):
        """Obtain the number of absolute command queues for force control

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return message
        """
        command = "froce_GetCount(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    #三指#
    def hand_get_firmware(self, ID):
        """Obtain the main version number of the three fingers

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return major version value
        """
        command = "Hand_GetFirmware(" + str(ID) + ")\n"
        value = int(self.send_command(command))
        if value == 255:
            return value
        return (int(self.send_command(command)))/10
    
    def hand_get_modified(self, ID):
        """Obtain the updated version number of the three fingers

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return minor version
        """
        command = "Hand_GetModified(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def hand_set_id(self, ID, value):
        """Set three finger machine ID

        Args:
            ID (int): Robot arm ID.
            Value (int): hand id value (1~254).
        
        Returns:
            str: return message
        """
        if value < 1 or value > 254:
            raise MyCobot630ProDataException(
            "The Hand  value must be 1 ~ 254, but received {}".format(value))
        command = "Hand_SetId(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def hand_get_id(self, ID):
        """Obtain the three finger machine ID

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return id value
        """
        command = "Hand_GetId(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def hand_set_enabled(self, ID, value):
        """Claw enable

        Args:
            ID (int): Robot arm ID.
            Value (int): Enable value (0 or 1).
        
        Returns:
            str: return message
        """
        if value < 0 or value > 1:
            raise MyCobot630ProDataException(
            "The SetEnabled  value must be 0 ~ 1, but received {}".format(value))
        command = "Hand_SetEnabled(" + str(ID) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def hand_set_joint_angle(self, ID, Jiont, value):
            """Set the angle of a single joint.

            Args:
                ID (int): Robot arm ID.
                Joint (int): Joint number (1~6).
                Value (int): Angle value (0~100).
            
            Returns:
                str: return message
            """
            if value < 0 or value > 100:
                raise MyCobot630ProDataException(
            "The Angle value must be 0 ~ 100, but received {}".format(value))
            if Jiont < 1 or Jiont > 6:
                raise MyCobot630ProDataException(
            "The Hand Jiont value must be 1 ~ 6, but received {}".format(Jiont))
            command = "Hand_SetJointAngle(" + str(ID) + "," + str(Jiont) + "," + str(value) + ")\n"
            return int(self.send_command(command))
    
    def hand_get_joint_angle(self, ID, Jiont):
        """Obtain the angle of a single joint.

        Args:
            ID (int): Robot arm ID.
            Joint (int): Joint number (1~6).
        
        Returns:
            str: return Single joint angle value
        """
        if Jiont < 1 or Jiont > 6:
            raise MyCobot630ProDataException(
            "The gripper Jiont value must be 1 ~ 6, but received {}".format(Jiont))
        command = "Hand_GetJointAngle(" + str(ID) + "," + str(Jiont) + ")\n"
        return int(self.send_command(command))
    
    
    def hand_set_joint_calibrate(self, ID, Jiont):
        """Initialize joints

        Args:
            ID (int): Robot arm ID.
            Joint (int): Joint number (1~6).
        
        Returns:
            str: return message
        """
        if Jiont < 1 or Jiont > 6:
            raise MyCobot630ProDataException(
            "The Hand Jiont value must be 1 ~ 6, but received {}".format(Jiont))
        command = "Hand_SetJointCalibrate(" + str(ID) + "," + str(Jiont) + ")\n"
        return int(self.send_command(command))
    

    def hand_get_state(self, ID):
        """Obtain the motion status of the three fingers.

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return Numerical value of motion state
        """
        command = "Hand_GetHand(" + str(ID) + ")\n"
        return int(self.send_command(command))
    
    def hand_set_torque(self, ID, Jiont, value):
        """Set the torque of the three finger joint.

        Args:
            ID (int): Robot arm ID.
            Joint (int): Joint number (1~6).
            Value (int): Torque value (0~100).
        
        Returns:
            str: return message
        """
        if value < 0 or value > 100:
                raise MyCobot630ProDataException(
            "The hand torque value must be 0 ~ 100, but received {}".format(value))
        if Jiont < 1 or Jiont > 6:
            raise MyCobot630ProDataException(
            "The Hand Jiont value must be 1 ~ 6, but received {}".format(Jiont))
        command = "Hand_SetTorque(" + str(ID) + "," + str(Jiont) + "," + str(value) + ")\n"
        return int(self.send_command(command))
    
    def hand_get_torque(self, ID, Jiont):
        """Obtain the torque of the three finger joint.

        Args:
            ID (int): Robot arm ID.
            Joint (int): Joint number (1~6).
        
        Returns:
            str: return Single Joint torque value
        """
        if Jiont < 1 or Jiont > 6:
            raise MyCobot630ProDataException(
            "The Hand Jiont value must be 1 ~ 6, but received {}".format(Jiont))
        command = "Hand_GetTorque(" + str(ID) + "," + str(Jiont) + ")\n"
        return int(self.send_command(command))
    
   
    def hand_set_speed(self, ID, Jiont, value):
        """Set the speed of the three finger joint.

        Args:
            ID (int): Robot arm ID.
            Joint (int): Joint number (1~6).
            Value (int): Speed value (1~100).
        
        Returns:
            str: return message
        """
        if value < 1 or value > 100:
                raise MyCobot630ProDataException(
            "The Hand speed value must be 1 ~ 100, but received {}".format(value))
        if Jiont < 1 or Jiont > 6:
             raise MyCobot630ProDataException(
            "The Hand Jiont value must be 1 ~ 6, but received {}".format(Jiont))
        command = "Hand_SetSpeed(" + str(ID) + "," + str(Jiont) + "," + str(value) + ")\n"
        return int(self.send_command(command))

 
    def hand_get_speed(self, ID, Jiont):
        """Obtain the velocity of the three finger joint.

        Args:
            ID (int): Robot arm ID.
            Joint (int): Joint number (1~6).
        
        Returns:
            str: return Single Joint speed value
        """
        if Jiont < 1 or Jiont > 6:
             raise MyCobot630ProDataException(
            "The Hand Jiont value must be 1 ~ 6, but received {}".format(Jiont))
        command = "Hand_GetSpeed(" + str(ID) + "," + str(Jiont) + ")\n"
        return int(self.send_command(command))
    
    def hand_set_fullangles(self, ID, angles, speed):
        """Set the angle of the three finger joint.

        Args:
            ID (int): Robot arm ID.
            Angles (list of int): A list containing 6 angle values, each value range must be (0~100).
            Value (int): Speed value (1~100).
        
        Returns:
            str: return message
        """
        if speed < 1 or speed > 100:
                 raise MyCobot630ProDataException(
            "The Hand speed value must be 1 ~ 100, but received {}".format(speed))
        if len(angles) == 6:
            if any(x < 0 or x > 100 for x in angles):
                 raise MyCobot630ProDataException(
            "The FullAngles value must be 0 ~ 100, but received {}".format(angles))
            else:
                command = "Hand_SetFullAngles(" + str(ID) + "," + str(angles[0]) + "," \
                                        + str(angles[1]) + "," + str(angles[2]) + "," \
                                        + str(angles[3]) + "," + str(angles[4]) + "," \
                                        + str(angles[5])+ "," + str(speed) + ")\n"
                return int(self.send_command(command))
        else:
            raise MyCobot630ProDataException(
            "Enter the number of angles must be 6, but received {}".format(len(angles)))

    def hand_get_fullangles(self, ID):
        """Obtain the total joint angle of three fingers.

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return All Joint angle value
。
        """
        command = "Hand_GetFullAngles(" + str(ID) + ")\n"
        angle_list= parse_mixed_list_string(self.send_command(command))
        return angle_list

    def hand_set_catch(self, ID, pose, value, num=0):   
        """Three finger gesture control.

        Args:
            ID (int): Robot arm ID.
            Pose (int): gesture type (0~4).
            Value (int): threshold (0~5), when pose is 4, value is (0~20).
            Num (int): defaults to all hand movements when no value is passed in, and only moves pose type joints when a value is passed in.

        Returns:
            str: return message
        """
        if pose < 0 or pose > 4:
                raise MyCobot630ProDataException(
            "The Hand pose value must be 0 ~ 4, but received {}".format(pose))
        if pose == 4:
            if value < 1 or value > 20:
                raise MyCobot630ProDataException(
            "The pose is 4 so value must be 1 ~ 20, but received {}".format(value))
        else:
            if value < 0 or value > 5:
                raise MyCobot630ProDataException(
            "The pose is less than 4 value must be 0 ~ 5, but received {}".format(value))
        if num <= 0:
            command = "Hand_SetCatch(" + str(ID) + "," + str(pose) + "," + str(value) + ")\n"
        else :
            command = "Hand_SetCatch(" + str(ID) + "," + str(pose) + ","+ str(value) +  "," + str(num) +")\n"
        return int(self.send_command(command))
    
    def hand_get_model(self,ID):    
        """Obtain the type of three finger machine, such as left-hand 0 or right-hand 1.

        Args:
            ID (int): Robot arm ID.
        
        Returns:
            str: return finger machine type
        """
        command = "Hand_GetModel(" + str(ID) + ")\n"
        return int(self.send_command(command))


    #末端
    def get_end_firmware(self):             
        """Obtain the final major version number.

        Returns:
            str: return Atom major version number

        """
        command = "GetFirmwareEnd()\n"
        value = self.send_command(command)
        if value == "25.5":
            value = int((float(value)) * 10)
            return value
        value = float(value)
        return value

    def get_end_modify(self):                
        """Obtain the end update version number.

        Returns:
            str: return Atom update version number

        """
        command = "GetModifyEnd()\n"
        return int(self.send_command(command))
    def get_end_bt_status(self):                
        """Get the status of the end button.

        Returns:
            str: return message

        """
        command = "SetEndBtStatus()\n"
        return int(self.send_command(command))
    def set_end_color(self, red, green, blue):
        """Obtain the end color.

        Args:
            red (int): (0 ~ 255)。
            green (int): (0 ~ 255)。
            blue (int): (0 ~ 255)。
        
        Returns:
            str: return message
        """
        if red < 0 or green < 0 or blue < 0:
            raise MyCobot630ProDataException(
            "The input RGB parameter should be: 0~255, but received {},{},{}".format(red,green,blue))
        if red > 255 or green > 255 or blue > 255:
            raise MyCobot630ProDataException(
            "The input RGB parameter should be: 0~255, but received {},{},{}".format(red,green,blue))

        command = "SetLedColor(" + str(red) + "," + str(green) + "," + str(blue) + ")\n"
        return int(self.send_command(command))
    
        

if __name__ == "__main__":
    ep = ElephantRobot("192.168.1.248", 5001)
    resp = ep.start_client()
    if resp != True:
        print(resp)
        sys.exit(1)
    print(ep.wait(5))
    time.sleep(2)
    ep.force_get_angle(14,0)
    print(ep.Force_SetAngle(14,20))
    print(ep.Force_SetId(14,255))
    print(ep.get_coords())
    print(ep.get_speed())
    print(ep._power_on())
    print(ep._power_off())
    print(ep.check_running())
    print(ep.state_check())
    print(ep.program_open("a.tax"))
    print(ep.program_run(0))
    print(ep.read_next_error())
    print(ep.write_coords([1, 2, 3, 4, 5, 6], 110))
    print(ep.write_coord(1, 100, 200))
    print(ep.write_angles([10, 20, 30, 40, 50, 60], 110))
    print(ep.write_angle(3, 180, 200))
    print(ep.set_speed(377))
    print(ep.set_carte_torque_limit("x", 55))
    print(ep.set_upside_down(False))
    print(ep.set_payload(100))
    print(ep._state_on())
    print(ep._state_off())
    print(ep.task_stop())
    print(ep.jog_angle("j2", 1, 300))
    print(ep.jog_coord("rY", 0, 200))
    print(ep.get_digital_in(3))
    print(ep.get_digital_out(3))
    print(ep.set_digital_out(3, 1))
    print(ep.get_digital_out(3))
    print(ep.set_analog_out(1, 3.5))
    print(ep.get_acceleration())
    print(ep.set_acceleration(55))
    print(ep.command_wait_done())
    print(ep.get_variable("f"))
    print(ep.assign_variable("ss", '"eee"'))
    print(ep.get_joint_current(1))
    print(ep.wait(5))
    time.sleep(2)
    ep.force_get_angle(14,0)
    print(ep.Force_SetAngle(14,20))
    print(ep.Force_SetId(14,255))
    print(ep.get_coords())
    print(ep.get_speed())
    print(ep._power_on())
    print(ep._power_off())
    print(ep.check_running())
    print(ep.state_check())
    print(ep.program_open("a.tax"))
    print(ep.program_run(0))
    print(ep.read_next_error())
    print(ep.write_coords([1, 2, 3, 4, 5, 6], 110))
    print(ep.write_coord(1, 100, 200))
    print(ep.write_angles([10, 20, 30, 40, 50, 60], 110))
    print(ep.write_angle(3, 180, 200))
    print(ep.set_speed(377))
    print(ep.set_carte_torque_limit("x", 55))
    print(ep.set_upside_down(False))
    print(ep.set_payload(100))
    print(ep._state_on())
    print(ep._state_off())
    print(ep.task_stop())
    print(ep.jog_angle("j2", 1, 300))
    print(ep.jog_coord("rY", 0, 200))
    print(ep.get_digital_in(3))
    print(ep.get_digital_out(3))
    print(ep.set_digital_out(3, 1))
    print(ep.get_digital_out(3))
    print(ep.set_analog_out(1, 3.5))
    print(ep.get_acceleration())
    print(ep.set_acceleration(55))
    print(ep.command_wait_done())
    print(ep.get_variable("f"))
    print(ep.assign_variable("ss", '"eee"'))
    print(ep.get_joint_current(1))

    print(ep.set_gripper_mode(1))
    print(ep.wait(2))
    print(ep.set_gripper_mode(0))

    print(ep.set_gripper_value(100, 20))
    print(ep.wait(2))
    print(ep.set_gripper_value(50, 20))
    print(ep.wait(2))
    print(ep.set_gripper_value(0, 20))
    print(ep.wait(2))

    print(ep.set_gripper_enabled(1))
    print(ep.wait(2))
    print(ep.set_gripper_enabled(0))

    ep.stop_client()


    # name = input("input:")

    # print(name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
ultraArmP1.py

Python interface for the ultraArmP1 robotic arm.

Author: weijian.wang
Date: 2025-09-11
Description: None
"""

import threading
import time

from pymycobot.common import ProtocolCode


class UltraArmP1:
    """Class for controlling the ultraArmP1 robotic arm via serial communication.

    """

    def __init__(self, port, baudrate=115200, timeout=0.1, debug=False):
        """Initialize the ultraArmP1 robot communication.

        Args:
            port (str): Serial port name (e.g., 'COM3' or '/dev/ttyUSB0').
            baudrate (int, optional): Communication baud rate. Defaults to 115200.
            timeout (float, optional): Serial read timeout in seconds. Defaults to 0.1.
            debug (bool, optional): Whether to print debug information. Defaults to False.
        """
        import serial

        self._serial_port = serial.Serial()
        self._serial_port.port = port
        self._serial_port.baudrate = baudrate
        self._serial_port.timeout = timeout
        self._serial_port.rts = True
        self._serial_port.dtr = True
        self._serial_port.open()
        self.debug = debug
        self.lock = threading.Lock()
        time.sleep(1)

    def _respone(self, timeout=90, _async=True):
        """Wait for device response from the serial buffer.

        This method continuously reads data from the serial port until either:
          - A specific keyword ("timx") is detected in the incoming data, or
          - The timeout is reached.

        Args:
            timeout (float): Maximum time (in seconds) to wait for a valid response.
            _async (bool): Whether to wait for a response.

        Returns:
            str: 'ok' if the keyword response ("timx") is detected.
            bool: False if the timeout occurs without receiving a valid response.
        """

        import time
        if not _async:
            return 1
        start_time = time.time()
        received_data = b""
        while time.time() - start_time < timeout:
            received_data += self._serial_port.read(self._serial_port.inWaiting())
            # if b"end" in received_data.lower():
            #     return 'ok'
            end_count = received_data.lower().count(b"end")
            if end_count >= 2:
                return 'ok'
            time.sleep(0.02)
        # Timeout
        if self.debug:
            print("data (timeout):", received_data)

        return False

    def _request(self, flag=""):
        """Send a request and read data from the robot.

        Args:
            flag (str, optional): Type of data expected ('angle', 'coord', or None).

        Returns:
            list[float] or int: Parsed angles/coordinates, or -1 if failed.
        """
        raw_data = None
        attempt = 0

        while attempt < 5:
            if self._serial_port.inWaiting() > 0:
                raw_data = self._serial_port.read(self._serial_port.inWaiting()).decode()
                if self.debug:
                    print("\nReceived data:\n%s**********************\n" % raw_data)

                if "ERROR: COMMAND NOT RECOGNIZED" in raw_data:
                    flag = None

                if flag == "angle":
                    data_lower = raw_data.lower()
                    if "angles" in data_lower:
                        angle_str = data_lower[data_lower.find("angles"):]
                        start_idx = angle_str.find("[")
                        end_idx = angle_str.find("]")
                        try:
                            angles_list = list(map(float, angle_str[start_idx + 1:end_idx].split(",")))
                            angles_list = [round(a, 2) for a in angles_list]
                            return angles_list
                        except Exception:
                            print("Received angles is not completed! Retry receive...")
                            attempt += 1
                            continue
                    else:
                        return -1

                elif flag == "coord":
                    data_lower = raw_data.lower()
                    if "coords" in data_lower:
                        coord_str = data_lower[data_lower.find("coords"):]
                        start_idx = coord_str.find("[")
                        end_idx = coord_str.find("]")
                        try:
                            coords_list = list(map(float, coord_str[start_idx + 1:end_idx].split(",")))
                            coords_list = [round(c, 2) for c in coords_list]
                            return coords_list
                        except Exception:
                            print("Received coords is not completed! Retry receive...")
                            attempt += 1
                            continue
                    else:
                        return -1

                elif flag is None:
                    return -1

            attempt += 1
            time.sleep(0.01)

        return -1

    def _debug(self, data):
        """Print debug information if debug mode is enabled.

        Args:
            data (str): Command or data to print.
        """
        if self.debug:
            print("\n***** Debug Info *****\nsend command: %s" % data)

    def set_unlock(self):
        """Unlock the robot's safety state."""
        with self.lock:
            command = ProtocolCode.SET_UNLOCK + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone()

    def set_reboot(self):
        """Reboot the robot controller board."""
        with self.lock:
            command = ProtocolCode.SET_REBOOT + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone()

    def set_joint_disable(self):
        """Disable the robot joints."""
        with self.lock:
            command = ProtocolCode.SET_JOINT_DISABLE + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone()

    def set_joint_enable(self):
        """Enable the robot joints."""
        with self.lock:
            command = ProtocolCode.SET_JOINT_ENABLE + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone()

    def get_angles_info(self):
        """Get the current joint angles of the robot.

        Returns:
            list[float] or int: Joint angles [J1, J2, J3, J4] or -1 if failed.
        """
        with self.lock:
            command = ProtocolCode.GET_CURRENT_ANGLES_COORDS_INFO + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._request("angle")

    def get_coords_info(self):
        """Get the current Cartesian coordinates of the robot.

        Returns:
            list[float] or int: Coordinates [X, Y, Z, E] or -1 if failed.
        """
        with self.lock:
            command = ProtocolCode.GET_CURRENT_ANGLES_COORDS_INFO + ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._request("coord")

    def set_coords(self, coords, speed, _async=True):
        """Move the robot using Cartesian coordinate control.

        Args:
            coords (list[float]): Coordinates [X, Y, Z].
            speed (int): Movement speed (1~5700).
            _async: (bool): Closed-loop switch
        """
        with self.lock:
            command = ProtocolCode.SET_ANGLES_COORDS
            if len(coords) > 0 and coords[0] is not None:
                command += f" X{coords[0]}"
            if len(coords) > 1 and coords[1] is not None:
                command += f" Y{coords[1]}"
            if len(coords) > 2 and coords[2] is not None:
                command += f" Z{coords[2]}"
            if speed is not None and 1 <= speed <= 5700:
                command += f" F{speed}"

            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone(_async=_async)

    def set_angle(self, joint_id, angle, speed, _async=True):
        """Set a single joint angle.

        Args:
            joint_id (int): Joint number (1~4).
            angle (float): Angle value.
            speed (int): Movement speed (1~5700).
            _async: (bool): Closed-loop switch
        """
        with self.lock:
            command = ProtocolCode.SET_ANGLES_COORDS
            joint_map = {1: "A", 2: "B", 3: "C", 4: "D"}
            if joint_id in joint_map:
                command += f" {joint_map[joint_id]}{angle}"
            if speed > 0:
                command += f" F{speed}"
            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone(_async=_async)

    def set_angles(self, angles, speed, _async=True):
        """Move robot using joint angle control.

        Args:
            angles (list[float]): Joint angles [J1, J2, J3, J4].
            speed (int): Movement speed (1~5700).
            _async: (bool): Closed-loop switch
        """
        with self.lock:
            command = ProtocolCode.SET_ANGLES_COORDS
            if len(angles) > 0 and angles[0] is not None:
                command += f" A{angles[0]}"
            if len(angles) > 1 and angles[1] is not None:
                command += f" B{angles[1]}"
            if len(angles) > 2 and angles[2] is not None:
                command += f" C{angles[2]}"
            if len(angles) > 3 and angles[3] is not None:
                command += f" D{angles[3]}"
            if speed is not None and 1 <= speed <= 5700:
                command += f" F{speed}"

            command += ProtocolCode.END
            self._serial_port.write(command.encode())
            self._serial_port.flush()
            self._debug(command)
            return self._respone(_async=_async)

    def close(self):
        """Close the serial port."""
        with self.lock:
            self._serial_port.close()

    def open(self):
        """Open the serial port."""
        with self.lock:
            self._serial_port.open()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import json
import struct
import socket
import sys
import threading
import time
import traceback
from pymycobot.mercury_api import MercuryCommandGenerator
from pymycobot.error import calibration_parameters

"""
for mercury x1 robot arms (six-axis or seven-axis), using socket to communicate with the robot
"""

SYS_VERSION_INFO = sys.version_info.major


def format_hex_log(data):
    if SYS_VERSION_INFO == 2:
        command_log = ""
        for d in data:
            command_log += hex(ord(d))[2:] + " "
    else:
        command_log = ""
        for d in data:
            command_log += hex(d)[2:] + " "

    return command_log


class MercuryArmsSocket(MercuryCommandGenerator):
    def __init__(self, arm, ip, netport=9000, debug=False):
        """
        Arm socket connection
        Args:
            arm: 'left_arm' or 'right_arm'
            ip: ip address
            netport: port
            debug: debug mode
        """
        super(MercuryArmsSocket, self).__init__(debug)
        self.crc_robot_class.append(self.__class__.__name__)
        self.arm = arm
        self.calibration_parameters = calibration_parameters
        self.SERVER_IP = ip
        self.SERVER_PORT = netport
        self.sock = self.connect_socket()

        self.read_threading = threading.Thread(target=self.read_thread, args=("socket",))
        self.read_threading.daemon = True
        self.read_threading.start()
        self.get_limit_switch()

    def read_thread(self, method=None):
        self.sock.settimeout(3)
        while True:
            try:
                data = self.sock.recv(1024)
                result = self._process_received(data)

                self.log.debug(f"_read :{format_hex_log(result)}")
                if not result:
                    continue

                with self.lock:
                    self.read_command.append([result, time.time()])

            except socket.timeout:
                time.sleep(0.1)

            except Exception as e:
                print(e)
                traceback.print_exc()

    def connect_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.SERVER_IP, self.SERVER_PORT))
        return sock

    def open(self):
        self.sock = self.connect_socket()

    def close(self):
        self.sock.close()
        self.sock = None

    def __format(self, command: str) -> bytes:
        send_data = {"command": command, "arm": self.arm}
        data_byter = json.dumps(send_data).encode('utf-8')
        date_length = struct.pack('!I', len(data_byter))
        return b''.join([date_length, data_byter])

    def _write(self, command, method=None):
        self.log.debug("_write: {}".format(format_hex_log(command)))
        self.sock.sendall(self.__format(command))

    def _send_command(self, genre, real_command):
        self.write_command.append(genre)
        self._write(self._flatten(real_command), method="socket")


if __name__ == '__main__':
    mercury_socket = MercuryArmsSocket("left_arm", ip="192.168.1.216", debug=True)
    mercury_socket.send_angle(1, 20, 10)
    # while True:
    #     print(mercury_socket.get_angles())
    #     time.sleep(0.1)



# coding=utf-8
import socket
import sys
import time
import traceback

from pymycobot.pro630 import Pro630Api
from pymycobot.common import ProtocolCode

SYS_VERSION_INFO = sys.version_info.major


def format_hex_log(data):
    if SYS_VERSION_INFO == 2:
        command_log = ""
        for d in data:
            command_log += hex(ord(d))[2:] + " "
    else:
        command_log = ""
        for d in data:
            command_log += hex(d)[2:] + " "

    return command_log


def connect_socket(addr, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((addr, port))
    return sock


class Pro630Client(Pro630Api):
    def __init__(self, host, port=9000, timeout=0.1, debug=False):
        """
        Args:
            host: host address
            port: port number
            timeout: socket timeout
            debug: debug mode
        """
        self.sock = connect_socket(host, port)
        self.timeout = timeout
        self.host = host
        self.port = port
        super().__init__(debug=debug, method='socket')

    def close(self):
        self.sock.close()

    def read_thread(self, method=None):
        self.sock.settimeout(self.timeout)
        while True:
            try:
                data = self.sock.recv(1024)
                print(data)
                result = self._process_received(data)
                command_log = format_hex_log(result)
                self.log.debug(f"_read :{command_log}")
                if not result:
                    continue

                with self.lock:
                    self.read_command.append([result, time.time()])
            except socket.timeout:
                print("socket timeout")
                time.sleep(0.1)
            except Exception as e:
                print(e)
                traceback.print_exc()

    def set_basic_output(self, pin_no, pin_signal):
        """Set basic output.IO low-level output high-level, high-level output high resistance state

        Args:
            pin_no: pin port number. range 1 ~ 6
            pin_signal: 0 / 1
        """
        return self._mesg(ProtocolCode.SET_BASIC_OUTPUT, pin_no, pin_signal)

    def get_basic_input(self, pin_no):
        """Get basic input.

        Args:
            pin_no: pin port number. range 1 ~ 6
            
        Return:
            1 - high
            0 - low
        """
        return self._mesg(ProtocolCode.GET_BASIC_INPUT, pin_no)

    def send_angles_sync(self, angles, speed):
        self.calibration_parameters(class_name = self.__class__.__name__, angles=angles, speed=speed)
        angles = [self._angle2int(angle) for angle in angles]
        return self._mesg(ProtocolCode.SEND_ANGLES, angles, speed, no_return=True)

    def set_monitor_mode(self, mode):
        raise NotImplementedError("Pro630 does not support monitor mode")

    def get_monitor_mode(self):
        raise NotImplementedError("Pro630 does not support monitor mode")


import socket
import threading
import serial

lock = threading.Lock()


def hex_to_signed_decimal(hex_str, bits=16):
    value = int(hex_str, 16)  # 先转换为无符号整数
    if value >= 2 ** (bits - 1):  # 如果超出正数范围，转换为负数
        value -= 2 ** bits
    return value


class Exoskeleton:
    def __init__(self, port):
        self.ser = serial.Serial(port=port, baudrate=1000000)
        self.dataLen = 0

    def _parse_data(self, data):
        parsed_data = []
        for i in range(7):
            data_h = data[0 + i * 4: 2 + i * 4]
            data_l = data[2 + i * 4: 4 + i * 4]
            encode = int(data_h + data_l, 16)
            angle = 0 if encode == 2048 else (
                180 * (encode - 2048) / 2048 if encode > 2048 else -180 * (2048 - encode) / 2048)
            parsed_data.append(round(angle, 2))

        button = bin(int(data[28: 30], 16))[2:].rjust(4, "0")
        parsed_data.extend([int(button[-4]), int(button[-1]), int(button[-3]), int(button[-2]), int(data[30: 32], 16),
                            int(data[32: 34], 16)])
        if self.dataLen == 31:  # 适配新固件带坐标信息
            for j in range(6):
                data_h = data[34 + j * 4: 36 + j * 4]
                data_l = data[36 + j * 4: 38 + j * 4]
                coord = hex_to_signed_decimal(data_h + data_l, 16)
                parsed_data.append(round(coord / 10, 2))
        return parsed_data

    def _parse_all_data(self, data):
        right_data = []
        left_data = []
        if self.dataLen == 60:  # 适配新固件带坐标信息
            for i in range(7):
                data_h = data[0 + i * 4: 2 + i * 4]
                data_l = data[2 + i * 4: 4 + i * 4]
                encode = int(data_h + data_l, 16)
                angle = 0 if encode == 2048 else (
                    180 * (encode - 2048) / 2048 if encode > 2048 else -180 * (2048 - encode) / 2048)
                left_data.append(round(angle, 2))

            button = bin(int(data[28: 30], 16))[2:].rjust(4, "0")
            left_data.extend([int(button[-4]), int(button[-1]), int(button[-3]), int(button[-2]), int(data[30: 32], 16),
                              int(data[32: 34], 16)])

            for i in range(7):
                data_h = data[34 + i * 4: 36 + i * 4]
                data_l = data[36 + i * 4: 38 + i * 4]
                encode = int(data_h + data_l, 16)
                angle = 0 if encode == 2048 else (
                    180 * (encode - 2048) / 2048 if encode > 2048 else -180 * (2048 - encode) / 2048)
                right_data.append(round(angle, 2))

            button = bin(int(data[62: 64], 16))[2:].rjust(4, "0")
            right_data.extend(
                [int(button[-4]), int(button[-1]), int(button[-3]), int(button[-2]), int(data[64: 66], 16),
                 int(data[66: 68], 16)])

            for j in range(6):
                data_h = data[68 + j * 4: 70 + j * 4]
                data_l = data[70 + j * 4: 72 + j * 4]
                coord = hex_to_signed_decimal(data_h + data_l, 16)
                left_data.append(round(coord / 10, 2))

            for j in range(6):
                data_h = data[92 + j * 4: 94 + j * 4]
                data_l = data[94 + j * 4: 96 + j * 4]
                coord = hex_to_signed_decimal(data_h + data_l, 16)
                right_data.append(round(coord / 10, 2))
        else:
            for i in range(7):
                data_h = data[0 + i * 4: 2 + i * 4]
                data_l = data[2 + i * 4: 4 + i * 4]
                encode = int(data_h + data_l, 16)
                angle = 0 if encode == 2048 else (
                    180 * (encode - 2048) / 2048 if encode > 2048 else -180 * (2048 - encode) / 2048)
                left_data.append(round(angle, 2))

            button = bin(int(data[28: 30], 16))[2:].rjust(4, "0")
            left_data.extend(
                [int(button[-4]), int(button[-1]), int(button[-3]), int(button[-2]), int(data[30: 32], 16),
                 int(data[32: 34], 16)])

            for i in range(7):
                data_h = data[34 + i * 4: 36 + i * 4]
                data_l = data[36 + i * 4: 38 + i * 4]
                encode = int(data_h + data_l, 16)
                angle = 0 if encode == 2048 else (
                    180 * (encode - 2048) / 2048 if encode > 2048 else -180 * (2048 - encode) / 2048)
                right_data.append(round(angle, 2))

            button = bin(int(data[62: 64], 16))[2:].rjust(4, "0")
            right_data.extend(
                [int(button[-4]), int(button[-1]), int(button[-3]), int(button[-2]), int(data[64: 66], 16),
                 int(data[66: 68], 16)])
        return [left_data, right_data]

    def _commmon(self, command_array):
        with lock:
            commmon_id = command_array[3]
            self.ser.write(command_array)
            start1 = self.ser.read().hex()
            if start1 != "fe" or self.ser.read().hex() != "fe":
                return None
            data_len = int(self.ser.read().hex(), 16)
            count = self.ser.in_waiting
            if data_len == count:
                self.dataLen = count
                data = self.ser.read(count).hex()
                if data[-2:] == "fa" and int(data[0: 2], 10) == commmon_id:
                    return data[2: -2]
        return None

    def get_all_data(self):
        get_all_array = [0xFE, 0xFE, 0x02, 0x01, 0xFA]
        data = self._commmon(get_all_array)
        if data is None:
            return None
        return self._parse_all_data(data)

    def get_arm_data(self, arm):
        if arm not in [1, 2]:
            raise ValueError("error arm")

        send_array = [0xFE, 0xFE, 0x03, 0x02, arm, 0xFA]
        data = self._commmon(send_array)
        if data is None:
            return None
        return self._parse_data(data)

    def get_joint_data(self, arm, id):
        if arm not in [1, 2] or id < 1 or id > 7:
            raise ValueError("error arm or id")

        send_array = [0xFE, 0xFE, 0x04, 0x03, arm, id, 0xFA]
        data = self._commmon(send_array)
        if data is None:
            return None
        encode = int(data[0: 2] + data[2: 4], 16)
        angle = 0 if encode == 2048 else (
            180 * (encode - 2048) / 2048 if encode > 2048 else -180 * (2048 - encode) / 2048)
        return round(angle, 2)

    def set_zero(self, arm, id):
        if arm not in [1, 2] or id < 1 or id > 7:
            raise ValueError("error arm or id")

        send_array = [0xFE, 0xFE, 0x04, 0x04, arm, id, 0xFA]
        return self._commmon(send_array)

    def set_color(self, arm, red, green, blue):
        if arm not in [1, 2]:
            raise ValueError("error arm")
        send_array = [0xFE, 0xFE, 0x06, 0x05, arm, red, green, blue, 0xFA]
        return self._commmon(send_array)
    
    def get_basic_version(self):
        send_array = [0xFE, 0xFE, 0x02, 0x06, 0xFA]
        return int(self._commmon(send_array), 16)/10
    
    def get_atom_version(self, arm):
        send_array = [0xFE, 0xFE, 0x03, 0x07, arm, 0xFA]
        return int(self._commmon(send_array)[2:], 16)/10

    def set_atom_matrix(self, arm):
        send_array = [0xFE, 0xFE, 0x03, 0x08, arm, 0xFA]
        return self._commmon(send_array)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import threading
import time

from .myagvapi import CommunicationProtocol, Utils, setup_serial_connect, setup_logging


class ProtocolCode:
    HEADER = (0xFE, 0xFE)
    RESTORE = (0x01, 0x00)
    SET_LED = (0x01, 0x02)
    SET_LED_MODE = (0x01, 0x0A)
    SET_MOTION_CONTROL = (0x01, 0x0B)
    GET_FIRMWARE_VERSION = (0x01, 0x03)
    GET_MODIFIED_VERSION = (0x01, 0x09)
    SET_GYRO_STATE = (0x01, 0x07)
    SET_AUTO_REPORT_STATE = (0x01, 0x0c)
    GET_AUTO_REPORT_STATE = (0x01, 0x0d)
    GET_GYRO_STATE = (0x01, 0x08)
    GET_MCU_INFO = ()


class MyAGVCommandProtocolApi(CommunicationProtocol):

    def __init__(self, debug=False):
        self.__movement = False
        self.__mutex = threading.Lock()
        self.log = setup_logging(name=f"{__name__}.{self.__class__.__name__}", debug=debug)

    def _read_command_buffer(self, timeout=1, conditional=None):
        previous_frame = b""
        is_record = False
        commands = b""

        start_time = time.time()
        while time.time() - start_time < timeout:
            current_frame = self.read()
            if current_frame == b'':
                time.sleep(0.001)
                continue

            if current_frame == b"\xfe" and previous_frame == b"\xfe" and is_record is False:
                is_record = True
                commands += b"\xfe\xfe"
                continue

            previous_frame = current_frame
            if is_record is False:
                continue

            commands += current_frame

            cond_res = True if conditional is None else conditional(commands)
            if sum(commands[2:-1]) & 0xff == commands[-1] and cond_res:
                break
        else:
            commands = b""
        return commands

    def _compose_complete_command(self, genre: ProtocolCode, params):  # packing command
        command_args = Utils.process_data_command(params)
        command_args = Utils.flatten(command_args)

        command = [*ProtocolCode.HEADER]
        if isinstance(genre, tuple):
            command.extend(genre)
        else:
            command.append(genre)

        command.extend(command_args)
        command.append(sum(command[2:]) & 0xff)
        return command

    def _parse_reply_instruction(self, genre, reply_data):  # unpacking command
        if len(reply_data) == 0:
            return None

        if genre == ProtocolCode.GET_FIRMWARE_VERSION:
            return Utils.float(reply_data[4], 1)

        elif genre == ProtocolCode.GET_MCU_INFO:
            index = 0
            res = []
            datas = reply_data[2:][:-1]  # header and footer frames are not counted
            while index < len(datas):
                if index in range(0, 3):
                    res.append(datas[index])
                    index += 1

                elif index in range(3, 15, 2):
                    data = Utils.decode_int16(datas[index:index + 2])
                    res.append(Utils.float(data, 2))
                    index += 2

                elif index == 15:
                    binary = bin(datas[index])[2:]
                    binary = binary.zfill(6)
                    res.append(binary)
                    index += 1

                elif index in (16, 17):
                    res.append(datas[index] / 10)
                    index += 1

                elif index in range(18, 26, 2):
                    res.append(Utils.float(Utils.decode_int16(datas[index:index + 2]), 3))
                    index += 2

                elif index in range(26, 32, 2):
                    res.append(Utils.float(Utils.decode_int16(datas[index:index + 2]), 3))
                    index += 2

                elif index in range(32, 42, 1):
                    res.append(datas[index])
                    index += 1
                else:
                    index += 1

            return res
        return reply_data[4]

    def _merge(self, genre, *args, has_reply=False):
        with self.__mutex:
            self.clear()
            if genre not in (ProtocolCode.GET_MCU_INFO,):
                real_command = self._compose_complete_command(genre, args)
                self.log.info(f"write: {' '.join(f'{x:02x}' for x in real_command)}")
                self.write(real_command)
                if has_reply is False:
                    return None

                for _ in range(8):
                    reply_data = self._read_command_buffer(conditional=lambda x: len(x) > 5, timeout=1)
                    if len(reply_data) == 0:
                        time.sleep(0.01)
                        continue

                    if reply_data[2] == genre[0] and reply_data[3] == genre[1]:
                        break

                    time.sleep(0.01)

                else:
                    reply_data = b""
            else:
                reply_data = self._read_command_buffer(conditional=lambda x: len(x) in (45, 29), timeout=1)

            self.log.info(f" read: {' '.join(f'{x:02x}' for x in reply_data)}")
            return self._parse_reply_instruction(genre, reply_data)


class MyAGVCommandApi(MyAGVCommandProtocolApi):
    def set_led(self, mode, r, g, b):
        """Set up LED lights

        Args:
            mode (int):
                1 - Set LED light color.
                2 - Set the LED light to blink
            r (int): 0 ~ 255
            g (int): 0 ~ 255
            b (int): 0 ~ 255
        """
        if mode not in (1, 2):
            raise ValueError("mode must be 1 or 2")

        if r not in range(256):
            raise ValueError("r must be 0 ~ 255")

        if g not in range(256):
            raise ValueError("g must be 0 ~ 255")

        if b not in range(256):
            raise ValueError("b must be 0 ~ 255")

        return self._merge(ProtocolCode.SET_LED, mode, r, g, b)

    def set_led_mode(self, mode: int):
        """Set the LED light mode

        Args:
            mode (int): 0 - charging mode, 1 - DIY mode
        """
        if mode not in [0, 1]:
            raise ValueError("mode must be 0 or 1")
        return self._merge(ProtocolCode.SET_LED_MODE, mode)

    def __basic_move_control(self, *genre, timeout: int = 5):
        t = time.time()
        self.__movement = True
        while time.time() - t < timeout:
            if self.__movement is False:
                break
            self._merge(ProtocolCode.SET_MOTION_CONTROL, *genre)
            time.sleep(0.1)
        self.stop()

    def go_ahead(self, speed: int, timeout: int = 5):
        """
        Control the car to move forward.
        Send control commands every 100ms.
        with a default motion time of 5 seconds.

        Args:
            speed (int): 1 ~ 127 is forward.
            timeout (int): default 5 s.
        """
        if not (0 < speed < 128):
            raise ValueError("speed must be between 1 and 127")
        self.__basic_move_control(128 + speed, 128, 128, timeout=timeout)

    def retreat(self, speed, timeout=5):
        """
        Control the car back. Send control commands every 100ms.
        with a default motion time of 5 seconds

        Args:
            speed (int): 1 ~ 127 is backward
            timeout (int): default 5 s.
        """
        if not (0 < speed < 128):
            raise ValueError("back_speed must be between 1 and 127")
        self.__basic_move_control(128 - speed, 128, 128, timeout=timeout)

    def pan_left(self, speed, timeout=5):
        """
        Control the car to pan to the left.
        Send control commands every 100ms. with a default motion time of 5 seconds

        Args:
            speed (int): 1 ~ 127
            timeout (int): default 5 s.
        """
        if not (0 < speed < 128):
            raise ValueError("pan_left_speed must be between 1 and 127")
        self.__basic_move_control(128, 128 + speed, 128, timeout=timeout)

    def pan_right(self, speed: int, timeout=5):
        """
        Control the car to pan to the right.
        Send control commands every 100ms. with a default motion time of 5 seconds

        Args:
            speed (int): 1 ~ 127
            timeout (int): default 5 s.
        """
        if not (0 < speed < 128):
            raise ValueError("pan_right_speed must be between 1 and 127")
        self.__basic_move_control(128, 128 - speed, 128, timeout=timeout)

    def clockwise_rotation(self, speed: int, timeout=5):
        """
        Control the car to rotate clockwise.
        Send control commands every 100ms. with a default motion time of 5 seconds

        Args:
            speed (int): 1 ~ 127
            timeout (int): default 5 s.
        """
        if speed < 1 or speed > 127:
            raise ValueError("speed must be between 1 and 127")
        self.__basic_move_control(128, 128, 128 - speed, timeout=timeout)

    def counterclockwise_rotation(self, speed: int, timeout=5):
        """
        Control the car to rotate counterclockwise.
        Send control commands every 100ms. with a default motion time of 5 seconds
        Args:
            speed (int): 1 ~ 127
            timeout (int): default 5 s.
        """
        if speed < 1 or speed > 127:
            raise ValueError("speed must be between 1 and 127")
        self.__basic_move_control(128, 128, 128 + speed, timeout=timeout)

    def stop(self):
        """stop-motion"""
        self._merge(ProtocolCode.SET_MOTION_CONTROL, 128, 128, 128)
        self.__movement = False

    def get_mcu_info(self):
        """
        Get MCU information
        Returns:
            MCU Version(list):
                version 1.0:
                    0-3: vx, vy, vz,
                    3-6: ax, ay, az,
                    6-9: wx, wy, wz,
                    10: battery status[str](
                        0-1: battery 2/1 interface access
                        2: The power adapter is plugged in
                        3: Plug in the charging pile
                        4-5: battery 1/2 LED status
                        ),
                    11-13: batter1/2 voltage,
                    13-15: battery1/2 current
                version 1.1:
                    15-18: Angle range (rollover, pitch, yaw)
                    18: battery status 1 - normal, 0 - abnormal
                    19: MPU status 1 - normal, 0 - abnormal
                    20-24: stall state 1 - stalled, 0 - normal
                    24-28: encoder status 1 - normal, 0 - abnormal
        """
        return self._merge(ProtocolCode.GET_MCU_INFO, has_reply=True)

    def restore(self):
        """Motor stall recovery"""
        self._merge(ProtocolCode.RESTORE, 0x01)

    def set_gyro_state(self, state=1):
        """Set gyroscope calibration status (save after power failure)

        Args:
            state (int, optional): 1 - open. 0 - close. Defaults to 1.
        """
        if state not in (0, 1):
            raise ValueError("state must be 0 or 1")
        self._merge(ProtocolCode.SET_GYRO_STATE, state)

    def get_gyro_state(self):
        """Get gyroscope calibration status

        Return:
            1 - open
            0 - close
        """
        return self._merge(ProtocolCode.GET_GYRO_STATE, has_reply=True)

    def get_firmware_version(self):
        """Get firmware version number"""
        return self._merge(ProtocolCode.GET_FIRMWARE_VERSION, has_reply=True)

    def get_modified_version(self):
        """Get modified version number"""
        return self._merge(ProtocolCode.GET_MODIFIED_VERSION, has_reply=True)

    def set_auto_report_state(self, state):
        """Set the state of automatic reporting

        Args:
            state (int): 1 - open. 0 - close. Defaults to 0.
        """
        if state not in (0, 1):
            raise ValueError("state must be 0 or 1")

        self._merge(ProtocolCode.SET_AUTO_REPORT_STATE, state)

    def get_auto_report_state(self):
        """Get the state of automatic reporting

        Return:
            1 - open
            0 - close
        """

        return self._merge(ProtocolCode.GET_AUTO_REPORT_STATE, has_reply=True)


class MyAgv(MyAGVCommandApi):

    def __init__(self, comport, baudrate=115200, timeout=0.1, debug=False):
        super().__init__(debug=debug)
        self._serial_port = setup_serial_connect(comport, baudrate, timeout)

    def write(self, command):
        self._serial_port.write(bytearray(command))

    def read(self, size=1):
        return self._serial_port.read(size)

    def close(self):
        if self._serial_port.is_open:
            self._serial_port.close()

    def open(self):
        if not self._serial_port.is_open:
            self._serial_port.open()

    def is_open(self):
        return self._serial_port.is_open

    def clear(self):
        self._serial_port.reset_output_buffer()
        self._serial_port.reset_input_buffer()
